# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import namedtuple
from typing import List, Optional, Type

import tensorflow as tf

import nncf
from nncf.common.graph.operator_metatypes import INPUT_NOOP_METATYPES
from nncf.common.graph.operator_metatypes import NOOP_METATYPES
from nncf.common.graph.operator_metatypes import OperatorMetatype
from nncf.common.graph.operator_metatypes import OperatorMetatypeRegistry
from nncf.common.hardware.opset import HWConfigOpName
from nncf.tensorflow.graph.metatypes.tf_ops import TF_OPERATION_METATYPES

WeightDef = namedtuple("WeightDef", ["weight_attr_name", "channel_axes"])

KERAS_LAYER_METATYPES = OperatorMetatypeRegistry("keras_layer_metatypes")


class TFLayerMetatype(OperatorMetatype):
    keras_layer_names: List[str] = []
    subtypes: List[Type[OperatorMetatype]] = []

    @classmethod
    def get_subtypes(cls) -> List[Type[OperatorMetatype]]:
        return cls.subtypes

    @classmethod
    def determine_subtype(cls, layer: tf.keras.layers.Layer) -> Optional[Type[OperatorMetatype]]:
        return cls._determine_subtype(layer)

    @classmethod
    def determine_subtype_wrapped_layer(
        cls, layer: tf.keras.layers.Layer, wrapper: Optional[tf.keras.layers.Wrapper] = None
    ) -> Optional[Type[OperatorMetatype]]:
        return cls._determine_subtype(layer, wrapper)

    @classmethod
    def get_all_aliases(cls) -> List[str]:
        return cls.keras_layer_names

    @classmethod
    def matches(cls, layer: tf.keras.layers.Layer, _: Optional[tf.keras.layers.Wrapper] = None) -> bool:
        return layer.__class__.__name__ in cls.keras_layer_names

    @classmethod
    def _determine_subtype(
        cls, layer: tf.keras.layers.Layer, wrapper: Optional[tf.keras.layers.Wrapper] = None
    ) -> Optional[Type[OperatorMetatype]]:
        matches = []
        for subtype in cls.get_subtypes():
            if subtype.matches(layer, wrapper):
                subtype_matches = subtype.determine_subtype_wrapped_layer(layer, wrapper)
                if subtype_matches is not None:
                    matches.extend(subtype_matches)
                else:
                    matches.append(subtype)
        if len(matches) > 1:
            raise nncf.InternalError("Multiple subtypes match operator call - cannot determine single subtype.")
        if not matches:
            return None
        return matches[0]


class TFLayerWithWeightsMetatype(TFLayerMetatype):
    weight_definitions: List[WeightDef] = []
    bias_attr_name: Optional[str] = None


@KERAS_LAYER_METATYPES.register()
@NOOP_METATYPES.register()
class TFLayerNoopMetatype(TFLayerMetatype):
    name = "noop"

    @classmethod
    def get_all_aliases(cls) -> List[str]:
        return [cls.name]


@KERAS_LAYER_METATYPES.register(is_subtype=True)
class TFDepthwiseConv1DSubLayerMetatype(TFLayerWithWeightsMetatype):
    name = "DepthwiseConv1D(Conv1DKerasLayer)"
    keras_layer_names = ["Conv1D", "Convolution1D"]
    hw_config_names = [HWConfigOpName.DEPTHWISECONVOLUTION]

    weight_definitions = [WeightDef(weight_attr_name="kernel", channel_axes=-1)]
    bias_attr_name = "bias"

    @classmethod
    def matches(cls, layer: tf.keras.layers.Layer, wrapper: Optional[tf.keras.layers.Wrapper] = None) -> bool:
        return layer.__class__.__name__ in cls.keras_layer_names and _is_depthwise_conv(layer, wrapper)


@KERAS_LAYER_METATYPES.register()
class TFConv1DLayerMetatype(TFLayerWithWeightsMetatype):
    name = "Conv1DKerasLayer"
    keras_layer_names = ["Conv1D", "Convolution1D"]
    hw_config_names = [HWConfigOpName.CONVOLUTION]
    subtypes = [TFDepthwiseConv1DSubLayerMetatype]

    weight_definitions = [WeightDef(weight_attr_name="kernel", channel_axes=-1)]
    bias_attr_name = "bias"


@KERAS_LAYER_METATYPES.register(is_subtype=True)
class TFDepthwiseConv2DSubLayerMetatype(TFLayerWithWeightsMetatype):
    name = "DepthwiseConv2D(Conv2DKerasLayer)"
    keras_layer_names = ["Conv2D", "Convolution2D"]
    hw_config_names = [HWConfigOpName.DEPTHWISECONVOLUTION]

    weight_definitions = [WeightDef(weight_attr_name="kernel", channel_axes=-1)]
    bias_attr_name = "bias"

    @classmethod
    def matches(cls, layer: tf.keras.layers.Layer, wrapper: Optional[tf.keras.layers.Wrapper] = None) -> bool:
        return layer.__class__.__name__ in cls.keras_layer_names and _is_depthwise_conv(layer, wrapper)


@KERAS_LAYER_METATYPES.register()
class TFConv2DLayerMetatype(TFLayerWithWeightsMetatype):
    name = "Conv2DKerasLayer"
    keras_layer_names = ["Conv2D", "Convolution2D"]
    hw_config_names = [HWConfigOpName.CONVOLUTION]
    subtypes = [TFDepthwiseConv2DSubLayerMetatype]

    weight_definitions = [WeightDef(weight_attr_name="kernel", channel_axes=-1)]
    bias_attr_name = "bias"


@KERAS_LAYER_METATYPES.register(is_subtype=True)
class TFDepthwiseConv3DSubLayerMetatype(TFLayerWithWeightsMetatype):
    name = "DepthwiseConv3D(Conv3DKerasLayer)"
    keras_layer_names = ["Conv3D", "Convolution3D"]
    hw_config_names = [HWConfigOpName.DEPTHWISECONVOLUTION]

    weight_definitions = [WeightDef(weight_attr_name="kernel", channel_axes=-1)]
    bias_attr_name = "bias"

    @classmethod
    def matches(cls, layer: tf.keras.layers.Layer, wrapper: Optional[tf.keras.layers.Wrapper] = None) -> bool:
        return layer.__class__.__name__ in cls.keras_layer_names and _is_depthwise_conv(layer, wrapper)


@KERAS_LAYER_METATYPES.register()
class TFConv3DLayerMetatype(TFLayerWithWeightsMetatype):
    name = "Conv3DKerasLayer"
    keras_layer_names = ["Conv3D", "Convolution3D"]
    hw_config_names = [HWConfigOpName.CONVOLUTION]
    subtypes = [TFDepthwiseConv3DSubLayerMetatype]

    weight_definitions = [WeightDef(weight_attr_name="kernel", channel_axes=-1)]
    bias_attr_name = "bias"


@KERAS_LAYER_METATYPES.register()
class TFDepthwiseConv2DLayerMetatype(TFLayerWithWeightsMetatype):
    name = "DepthwiseConv2DKerasLayer"
    keras_layer_names = ["DepthwiseConv2D"]
    hw_config_names = [HWConfigOpName.DEPTHWISECONVOLUTION]

    weight_definitions = [WeightDef(weight_attr_name="depthwise_kernel", channel_axes=(2, 3))]
    bias_attr_name = "bias"


@KERAS_LAYER_METATYPES.register()
class TFConv1DTransposeLayerMetatype(TFLayerWithWeightsMetatype):
    name = "Conv1DTransposeKerasLayer"
    keras_layer_names = ["Conv1DTranspose", "Convolution1DTranspose"]
    hw_config_names = [HWConfigOpName.CONVOLUTION]

    weight_definitions = [WeightDef(weight_attr_name="kernel", channel_axes=-2)]
    bias_attr_name = "bias"


@KERAS_LAYER_METATYPES.register()
class TFConv2DTransposeLayerMetatype(TFLayerWithWeightsMetatype):
    name = "Conv2DTransposeKerasLayer"
    keras_layer_names = ["Conv2DTranspose", "Convolution2DTranspose"]
    hw_config_names = [HWConfigOpName.CONVOLUTION]

    weight_definitions = [WeightDef(weight_attr_name="kernel", channel_axes=-2)]
    bias_attr_name = "bias"


@KERAS_LAYER_METATYPES.register()
class TFConv3DTransposeLayerMetatype(TFLayerWithWeightsMetatype):
    name = "Conv3DTransposeKerasLayer"
    keras_layer_names = ["Conv3DTranspose", "Convolution3DTranspose"]
    hw_config_names = [HWConfigOpName.CONVOLUTION]

    weight_definitions = [WeightDef(weight_attr_name="kernel", channel_axes=-2)]
    bias_attr_name = "bias"


@KERAS_LAYER_METATYPES.register()
class TFDenseLayerMetatype(TFLayerWithWeightsMetatype):
    name = "DenseKerasLayer"
    keras_layer_names = ["Dense"]
    hw_config_names = [HWConfigOpName.MATMUL]

    weight_definitions = [WeightDef(weight_attr_name="kernel", channel_axes=-1)]
    bias_attr_name = "bias"


@KERAS_LAYER_METATYPES.register()
class TFBatchNormalizationLayerMetatype(TFLayerWithWeightsMetatype):
    name = "BatchNormalizationKerasLayer"
    keras_layer_names = ["BatchNormalization", "SyncBatchNormalization"]

    weight_definitions = [WeightDef(weight_attr_name="gamma", channel_axes=0)]
    bias_attr_name = "beta"


@KERAS_LAYER_METATYPES.register()
class TFSeparableConv1DLayerMetatype(TFLayerWithWeightsMetatype):
    name = "SeparableConv1DKerasLayer"
    keras_layer_names = ["SeparableConv1D", "SeparableConvolution1D"]

    weight_definitions = [
        WeightDef(weight_attr_name="depthwise_kernel", channel_axes=(1, 2)),
        WeightDef(weight_attr_name="pointwise_kernel", channel_axes=-1),
    ]
    bias_attr_name = "bias"


@KERAS_LAYER_METATYPES.register()
class TFSeparableConv2DLayerMetatype(TFLayerWithWeightsMetatype):
    name = "SeparableConv2DKerasLayer"
    keras_layer_names = ["SeparableConv2D", "SeparableConvolution2D"]

    weight_definitions = [
        WeightDef(weight_attr_name="depthwise_kernel", channel_axes=(2, 3)),
        WeightDef(weight_attr_name="pointwise_kernel", channel_axes=-1),
    ]
    bias_attr_name = "bias"


@KERAS_LAYER_METATYPES.register()
class TFEmbeddingLayerMetatype(TFLayerWithWeightsMetatype):
    name = "EmbeddingKerasLayer"
    keras_layer_names = ["Embedding"]

    weight_definitions = [WeightDef(weight_attr_name="embeddings", channel_axes=None)]


@KERAS_LAYER_METATYPES.register()
class TFLocallyConnected1DLayerMetatype(TFLayerWithWeightsMetatype):
    name = "LocallyConnected1DKerasLayer"
    keras_layer_names = ["LocallyConnected1D"]

    weight_definitions = [WeightDef(weight_attr_name="kernel", channel_axes=None)]


@KERAS_LAYER_METATYPES.register()
class TFLocallyConnected2DLayerMetatype(TFLayerWithWeightsMetatype):
    name = "LocallyConnected2DKerasLayer"
    keras_layer_names = ["LocallyConnected2D"]

    weight_definitions = [WeightDef(weight_attr_name="kernel", channel_axes=None)]


@KERAS_LAYER_METATYPES.register()
class TFCropping1DLayerMetatype(TFLayerMetatype):
    name = "Cropping1DKerasLayer"
    keras_layer_names = ["Cropping1D"]
    hw_config_names = [HWConfigOpName.CROP]


@KERAS_LAYER_METATYPES.register()
class TFCropping2DLayerMetatype(TFLayerMetatype):
    name = "Cropping2DKerasLayer"
    keras_layer_names = ["Cropping2D"]
    hw_config_names = [HWConfigOpName.CROP]


@KERAS_LAYER_METATYPES.register()
class TFCropping3DLayerMetatype(TFLayerMetatype):
    name = "Cropping3DKerasLayer"
    keras_layer_names = ["Cropping3D"]
    hw_config_names = [HWConfigOpName.CROP]


@KERAS_LAYER_METATYPES.register()
class TFFlattenLayerMetatype(TFLayerMetatype):
    name = "FlattenKerasLayer"
    keras_layer_names = ["Flatten"]
    hw_config_names = [HWConfigOpName.FLATTEN]


@KERAS_LAYER_METATYPES.register()
class TFGlobalMaxPooling1DLayerMetatype(TFLayerMetatype):
    name = "GlobalMaxPooling1DKerasLayer"
    keras_layer_names = ["GlobalMaxPool1D", "GlobalMaxPooling1D"]
    hw_config_names = [HWConfigOpName.MAXPOOL]


@KERAS_LAYER_METATYPES.register()
class TFGlobalMaxPooling2DLayerMetatype(TFLayerMetatype):
    name = "GlobalMaxPooling2DKerasLayer"
    keras_layer_names = ["GlobalMaxPool2D", "GlobalMaxPooling2D"]
    hw_config_names = [HWConfigOpName.MAXPOOL]


@KERAS_LAYER_METATYPES.register()
class TFGlobalMaxPooling3DLayerMetatype(TFLayerMetatype):
    name = "GlobalMaxPooling3DKerasLayer"
    keras_layer_names = ["GlobalMaxPool3D", "GlobalMaxPooling3D"]
    hw_config_names = [HWConfigOpName.MAXPOOL]


@KERAS_LAYER_METATYPES.register()
class TFMaxPooling1DLayerMetatype(TFLayerMetatype):
    name = "MaxPooling1DKerasLayer"
    keras_layer_names = ["MaxPool1D", "MaxPooling1D"]
    hw_config_names = [HWConfigOpName.MAXPOOL]


@KERAS_LAYER_METATYPES.register()
class TFMaxPooling2DLayerMetatype(TFLayerMetatype):
    name = "MaxPooling2DKerasLayer"
    keras_layer_names = ["MaxPool2D", "MaxPooling2D"]
    hw_config_names = [HWConfigOpName.MAXPOOL]


@KERAS_LAYER_METATYPES.register()
class TFMaxPooling3DLayerMetatype(TFLayerMetatype):
    name = "MaxPooling3DKerasLayer"
    keras_layer_names = ["MaxPool3D", "MaxPooling3D"]
    hw_config_names = [HWConfigOpName.MAXPOOL]


@KERAS_LAYER_METATYPES.register()
class TFRepeatVectorLayerMetatype(TFLayerMetatype):
    name = "RepeatVectorKerasLayer"
    keras_layer_names = ["RepeatVector"]
    hw_config_names = [HWConfigOpName.TILE]


@KERAS_LAYER_METATYPES.register()
class TFReshapeLayerMetatype(TFLayerMetatype):
    name = "ReshapeKerasLayer"
    keras_layer_names = ["Reshape"]
    hw_config_names = [HWConfigOpName.RESHAPE]


@KERAS_LAYER_METATYPES.register()
class TFPermuteLayerMetatype(TFLayerMetatype):
    name = "PermuteKerasLayer"
    keras_layer_names = ["Permute"]
    hw_config_names = [HWConfigOpName.TRANSPOSE]


@KERAS_LAYER_METATYPES.register()
class TFZeroPadding1DLayerMetatype(TFLayerMetatype):
    name = "ZeroPadding1DKerasLayer"
    keras_layer_names = ["ZeroPadding1D"]
    hw_config_names = [HWConfigOpName.PAD]


@KERAS_LAYER_METATYPES.register()
class TFZeroPadding2DLayerMetatype(TFLayerMetatype):
    name = "ZeroPadding2DKerasLayer"
    keras_layer_names = ["ZeroPadding2D"]
    hw_config_names = [HWConfigOpName.PAD]


@KERAS_LAYER_METATYPES.register()
class TFZeroPadding3DLayerMetatype(TFLayerMetatype):
    name = "ZeroPadding3DKerasLayer"
    keras_layer_names = ["ZeroPadding3D"]
    hw_config_names = [HWConfigOpName.PAD]


@KERAS_LAYER_METATYPES.register()
class TFUpSampling1DLayerMetatype(TFLayerMetatype):
    # Split->Concat pattern
    name = "UpSampling1DKerasLayer"
    keras_layer_names = ["UpSampling1D"]


@KERAS_LAYER_METATYPES.register()
class TFUpSampling2DLayerMetatype(TFLayerMetatype):
    name = "UpSampling2DKerasLayer"
    keras_layer_names = ["UpSampling2D"]
    hw_config_names = [HWConfigOpName.INTERPOLATE]


@KERAS_LAYER_METATYPES.register()
class TFUpSampling3DLayerMetatype(TFLayerMetatype):
    name = "UpSampling3DKerasLayer"
    keras_layer_names = ["UpSampling3D"]
    hw_config_names = [HWConfigOpName.INTERPOLATE]


@KERAS_LAYER_METATYPES.register()
class TFConcatenateLayerMetatype(TFLayerMetatype):
    name = "ConcatenateKerasLayer"
    keras_layer_names = ["Concatenate"]
    hw_config_names = [HWConfigOpName.CONCAT]


@KERAS_LAYER_METATYPES.register()
class TFAddLayerMetatype(TFLayerMetatype):
    name = "AddKerasLayer"
    keras_layer_names = ["Add"]
    hw_config_names = [HWConfigOpName.ADD]


@KERAS_LAYER_METATYPES.register()
class TFSubtractLayerMetatype(TFLayerMetatype):
    name = "SubtractKerasLayer"
    keras_layer_names = ["Subtract"]
    hw_config_names = [HWConfigOpName.SUBTRACT]


@KERAS_LAYER_METATYPES.register()
class TFMultiplyLayerMetatype(TFLayerMetatype):
    name = "MultiplyKerasLayer"
    keras_layer_names = ["Multiply"]
    hw_config_names = [HWConfigOpName.MULTIPLY]


@KERAS_LAYER_METATYPES.register()
class TFRescalingLayerMetatype(TFLayerMetatype):
    name = "RescalingKerasLayer"
    keras_layer_names = ["Rescaling"]


@KERAS_LAYER_METATYPES.register()
class TFAveragePooling1DLayerMetatype(TFLayerMetatype):
    name = "AveragePooling1DKerasLayer"
    keras_layer_names = ["AveragePooling1D", "AvgPool1D"]
    hw_config_names = [HWConfigOpName.AVGPOOL]


@KERAS_LAYER_METATYPES.register()
class TFAveragePooling2DLayerMetatype(TFLayerMetatype):
    name = "AveragePooling2DKerasLayer"
    keras_layer_names = ["AveragePooling2D", "AvgPool2D"]
    hw_config_names = [HWConfigOpName.AVGPOOL]


@KERAS_LAYER_METATYPES.register()
class TFAveragePooling3DLayerMetatype(TFLayerMetatype):
    name = "AveragePooling3DKerasLayer"
    keras_layer_names = ["AveragePooling3D", "AvgPool3D"]
    hw_config_names = [HWConfigOpName.AVGPOOL]


@KERAS_LAYER_METATYPES.register()
class TFGlobalAveragePooling1DLayerMetatype(TFLayerMetatype):
    name = "GlobalAveragePooling1DKerasLayer"
    keras_layer_names = ["GlobalAveragePooling1D", "GlobalAvgPool1D"]
    hw_config_names = [HWConfigOpName.AVGPOOL]


@KERAS_LAYER_METATYPES.register()
class TFGlobalAveragePooling2DLayerMetatype(TFLayerMetatype):
    name = "GlobalAveragePooling2DKerasLayer"
    keras_layer_names = ["GlobalAveragePooling2D", "GlobalAvgPool2D"]
    hw_config_names = [HWConfigOpName.AVGPOOL]


@KERAS_LAYER_METATYPES.register()
class TFGlobalAveragePooling3DLayerMetatype(TFLayerMetatype):
    name = "GlobalAveragePooling3DKerasLayer"
    keras_layer_names = ["GlobalAveragePooling3D", "GlobalAvgPool3D"]
    hw_config_names = [HWConfigOpName.AVGPOOL]


@KERAS_LAYER_METATYPES.register()
class TFReLULayerMetatype(TFLayerMetatype):
    name = "ReLUKerasLayer"
    keras_layer_names = ["ReLU"]


@KERAS_LAYER_METATYPES.register()
class TFThresholdedReLULayerMetatype(TFLayerMetatype):
    name = "ThresholdedReLUKerasLayer"
    keras_layer_names = ["ThresholdedReLU"]


@KERAS_LAYER_METATYPES.register()
class TFELULayerMetatype(TFLayerMetatype):
    name = "ELUKerasLayer"
    keras_layer_names = ["ELU"]


@KERAS_LAYER_METATYPES.register()
class TFPReLULayerMetatype(TFLayerMetatype):
    name = "PReLUKerasLayer"
    keras_layer_names = ["PReLU"]


@KERAS_LAYER_METATYPES.register()
class TFLeakyReLULayerMetatype(TFLayerMetatype):
    name = "LeakyReLUKerasLayer"
    keras_layer_names = ["LeakyReLU"]


@KERAS_LAYER_METATYPES.register()
class TFSoftmaxLayerMetatype(TFLayerMetatype):
    name = "SoftmaxKerasLayer"
    keras_layer_names = ["Softmax"]


@KERAS_LAYER_METATYPES.register()
class TFActivationLayerMetatype(TFLayerMetatype):
    name = "ActivationKerasLayer"
    keras_layer_names = ["Activation"]


@KERAS_LAYER_METATYPES.register()
class TFAverageLayerMetatype(TFLayerMetatype):
    name = "AverageKerasLayer"
    keras_layer_names = ["Average"]


@KERAS_LAYER_METATYPES.register()
class TFLayerNormalizationLayerMetatype(TFLayerMetatype):
    name = "LayerNormalizationKerasLayer"
    keras_layer_names = ["LayerNormalization"]


@KERAS_LAYER_METATYPES.register()
@INPUT_NOOP_METATYPES.register()
class TFInputLayerMetatype(TFLayerMetatype):
    name = "InputLayer"
    keras_layer_names = ["InputLayer"]


@KERAS_LAYER_METATYPES.register()
class TFDropoutLayerMetatype(TFLayerMetatype):
    name = "DropoutKerasLayer"
    keras_layer_names = ["Dropout"]


@KERAS_LAYER_METATYPES.register()
class TFLambdaLayerMetatype(TFLayerMetatype):
    name = "LambdaKerasLayer"
    keras_layer_names = ["Lambda"]


@KERAS_LAYER_METATYPES.register()
class TFTensorFlowOpLayerMetatype(TFLayerMetatype):
    name = "TensorFlowOpKerasLayer"
    keras_layer_names = ["TensorFlowOpLayer"]

    @classmethod
    def get_subtypes(cls) -> List[Type[OperatorMetatype]]:
        return list(TF_OPERATION_METATYPES.registry_dict.values())

    @classmethod
    def determine_subtype(
        cls, layer: tf.keras.layers.Layer, wrapper: Optional[tf.keras.layers.Wrapper] = None
    ) -> Optional[Type[OperatorMetatype]]:
        return TF_OPERATION_METATYPES.get_operator_metatype_by_op_name(layer.node_def.op)


@KERAS_LAYER_METATYPES.register()
class TFOpLambdaMetatype(TFLayerMetatype):
    name = "TFOpLambdaKerasLayer"
    keras_layer_names = ["TFOpLambda"]

    @classmethod
    def get_subtypes(cls) -> List[Type[OperatorMetatype]]:
        return list(TF_OPERATION_METATYPES.registry_dict.values())

    @classmethod
    def determine_subtype(
        cls, layer: tf.keras.layers.Layer, wrapper: Optional[tf.keras.layers.Wrapper] = None
    ) -> Optional[Type[OperatorMetatype]]:
        return TF_OPERATION_METATYPES.get_operator_metatype_by_op_name(layer.symbol)


@KERAS_LAYER_METATYPES.register()
class TFSlicingOpLambdaMetatype(TFLayerMetatype):
    name = "SlicingOpLambdaKerasLayer"
    keras_layer_names = ["SlicingOpLambda"]

    @classmethod
    def get_subtypes(cls) -> List[Type[OperatorMetatype]]:
        return list(TF_OPERATION_METATYPES.registry_dict.values())

    @classmethod
    def determine_subtype(
        cls, layer: tf.keras.layers.Layer, wrapper: Optional[tf.keras.layers.Wrapper] = None
    ) -> Optional[Type[OperatorMetatype]]:
        return TF_OPERATION_METATYPES.get_operator_metatype_by_op_name(layer.symbol)


@KERAS_LAYER_METATYPES.register()
class TFNNCFWrapperLayerMetatype(TFLayerMetatype):
    name = "NNCFWrapperLayer"
    keras_layer_names = ["NNCFWrapper"]

    @classmethod
    def get_subtypes(cls) -> List[Type[OperatorMetatype]]:
        return list(KERAS_LAYER_METATYPES.registry_dict.values())

    @classmethod
    def determine_subtype(cls, layer: tf.keras.layers.Layer) -> Optional[Type[OperatorMetatype]]:
        unwrapped_layer = layer.layer
        unwrapped_layer_metatype = KERAS_LAYER_METATYPES.get_operator_metatype_by_op_name(
            unwrapped_layer.__class__.__name__
        )
        subtype = unwrapped_layer_metatype.determine_subtype_wrapped_layer(unwrapped_layer, layer)
        if subtype is not None:
            return subtype
        return unwrapped_layer_metatype

    @classmethod
    def _determine_subtype(
        cls, layer: tf.keras.layers.Layer, wrapper: Optional[tf.keras.layers.Wrapper] = None
    ) -> Optional[Type[OperatorMetatype]]:
        unwrapped_layer = layer.layer
        return super()._determine_subtype(unwrapped_layer, wrapper)


def _is_depthwise_conv(layer: tf.keras.layers.Layer, wrapper: Optional[tf.keras.layers.Wrapper] = None) -> bool:
    channel_axis = -1 - layer.rank if layer.data_format == "channels_first" else -1

    channels = (
        layer.get_input_shape_at(0)[channel_axis] if wrapper is None else wrapper.get_input_shape_at(0)[channel_axis]
    )

    if channels is None:
        raise ValueError("The channel dimension of the inputs should be defined. Found `None`.")

    input_channels = int(channels)

    return input_channels == layer.groups and input_channels > 1



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import ABC
from abc import abstractmethod
from itertools import islice
from typing import Any, Dict, Optional, TypeVar

import nncf
from nncf.common import factory
from nncf.common.graph.graph import NNCFGraph
from nncf.common.graph.transformations.layout import TransformationLayout
from nncf.common.logging.track_progress import track
from nncf.common.tensor import NNCFTensor
from nncf.common.tensor_statistics.statistic_point import StatisticPointsContainer
from nncf.data.dataset import Dataset

TensorType = TypeVar("TensorType")
TModel = TypeVar("TModel")

EMPTY_DATASET_ERROR = (
    "Calibration dataset must not be empty. Please provide calibration dataset with at least one sample."
)


class StatisticsAggregator(ABC):
    """
    Base class for statistics collection.
    """

    def __init__(self, dataset: Dataset):
        self.dataset = dataset
        self.stat_subset_size = None
        self.statistic_points = StatisticPointsContainer()

    def _get_iterations_number(self) -> Optional[int]:
        """
        Returns number of iterations.

        :return: Number of iterations for statistics collection.
        """
        dataset_length = self.dataset.get_length()
        if dataset_length and self.stat_subset_size:
            return min(dataset_length, self.stat_subset_size)
        return dataset_length or self.stat_subset_size

    def collect_statistics(self, model: TModel, graph: NNCFGraph) -> None:
        """
        Collects statistics for registered StatisticPoints.
        The statistics are stored in self.statistic_points.

        :param model: Backend-specific model instance.
        :param graph: Model graph.
        """
        if not self.statistic_points:
            return
        model_transformer = factory.ModelTransformerFactory.create(model)
        merged_statistics = self._get_merged_statistic_points(self.statistic_points, model, graph)
        transformation_layout = self._get_transformation_layout_extra_outputs(merged_statistics)
        model_with_outputs = model_transformer.transform(transformation_layout)
        engine = factory.EngineFactory.create(model_with_outputs)

        iterations_number = self._get_iterations_number()
        empty_statistics = True
        for input_data in track(
            islice(self.dataset.get_inference_data(), iterations_number),
            total=iterations_number,
            description="Statistics collection",
        ):
            outputs = engine.infer(input_data)
            processed_outputs = self._process_outputs(outputs)
            self._register_statistics(processed_outputs, merged_statistics)
            empty_statistics = False
        if empty_statistics:
            raise nncf.ValidationError(EMPTY_DATASET_ERROR)

    def register_statistic_points(self, statistic_points: StatisticPointsContainer) -> None:
        """
        Register statistic points for statistics collection and recalculates the maximum number samples
        for collecting statistics, based on the maximum value from the all algorithms.

        :param statistic_points: StatisticPointsContainer instance with the statistic points
        """
        for _, _statistic_points in statistic_points.items():
            for _statistic_point in _statistic_points:
                self.statistic_points.add_statistic_point(_statistic_point)

        for _, _statistic_points in self.statistic_points.items():
            for _statistic_point in _statistic_points:
                for _, tensor_collectors in _statistic_point.algorithm_to_tensor_collectors.items():
                    for tensor_collector in tensor_collectors:
                        if self.stat_subset_size is None:
                            self.stat_subset_size = tensor_collector.num_samples
                        elif tensor_collector.num_samples is not None:
                            self.stat_subset_size = max(self.stat_subset_size, tensor_collector.num_samples)

    @abstractmethod
    def _register_statistics(self, outputs: Dict[str, NNCFTensor], statistic_points: StatisticPointsContainer) -> None:
        """
        Process prepared raw model outputs and statistic points for the further usage.

        :param outputs: prepared raw model outputs
        :param statistic_points: StatisticPointsContainer instance with the statistic points
        """

    @abstractmethod
    def _get_transformation_layout_extra_outputs(
        self, statistic_points: StatisticPointsContainer
    ) -> TransformationLayout:
        """
        Creates backend-specific transformation layout for the further statistics collection.

        :param statistic_points: StatisticPointsContainer to add outputs
        :return: TransformationLayout with the corresponding transformations
        """

    @staticmethod
    @abstractmethod
    def _get_merged_statistic_points(
        statistic_points: StatisticPointsContainer, model: TModel, graph: NNCFGraph
    ) -> StatisticPointsContainer:
        """
        Creates a new StatisticPointContainer that has no duplicated tensor collectors for one
        unique statistic point. Alters statistic collectors in the given statistic point container so statistics
        collected by merged statistic collectors will be available in all corresponding statistic collectors
        from the given statistic point container.

        :param statistic_points: Registered statistic points with possible tensor collectors duplicates.
        :param model: Backend-specific target model.
        :param graph: Model graph.
        :return: Merged statistic points container bounded with given statistic point container.
        """

    @staticmethod
    @abstractmethod
    def _process_outputs(outputs: Any) -> Dict[str, NNCFTensor]:
        """
        Post-process model outputs for the further statistics collection.

        :param outputs: raw model outputs
        :return: processed model outputs in Dict[str, NNCFTensor] format
        """



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from abc import abstractmethod
from typing import Dict, List, Optional, TypeVar

from nncf.common.graph import NNCFGraph
from nncf.common.graph.transformations.commands import TargetPoint
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.tensor_statistics.collectors import TensorStatisticCollectorBase
from nncf.data.dataset import Dataset
from nncf.experimental.tensor.tensor import Tensor
from nncf.quantization.algorithms.layerwise.iterator import LayerwiseIterator
from nncf.quantization.algorithms.layerwise.scheduler import LayerwiseStep
from nncf.quantization.algorithms.layerwise.scheduler import NodeOutputPort

TModel = TypeVar("TModel")


class LayerwiseEngineBackend(ABC):
    @staticmethod
    @abstractmethod
    def create_layerwise_iterator(
        model: TModel,
        graph: NNCFGraph,
        schedule: List[LayerwiseStep],
        dataset: Dataset,
        subset_size: int = 100,
        cache: Optional[Dict[NodeOutputPort, List[Tensor]]] = None,
    ) -> LayerwiseIterator:
        """
        Creates an iterator to iterate through layers according to a given schedule.

        This method generates an iterator that processes each layer of the model sequentially,
        following the specified schedule. It allows for detailed examination and manipulation
        of the model's intermediate layer outputs.

        :param model: The model to be iterated over.
        :param graph: The model graph.
        :param schedule: A list of steps defining how each layer should be processed.
        :param dataset: The dataset to be used for obtaining inputs to the model.
        :param subset_size: The number of samples from the dataset to use, defaults to 100.
        :param cache: Optional cache to store and reuse layer outputs, defaults to None.
        :return: An iterator yielding specified inputs or outputs for each layer of the model.
        """

    @staticmethod
    @abstractmethod
    def target_point(target_type: TargetType, target_node_name: str, port_id: int) -> TargetPoint:
        """
        Returns backend-specific target point.

        :param target_type: Type of the location that should be modified.
        :param target_node_name: Name of the located node.
        :param port_id: id of the port for the statistics distribution.
        :return: Backend-specific TargetPoint.
        """

    @staticmethod
    @abstractmethod
    def raw_statistic_collector(num_samples: Optional[int] = None) -> TensorStatisticCollectorBase:
        """
        Returns backend-specific raw statistic collector.
        This statistic collector is used for raw data calculation, without aggregating.

        :param num_samples: Maximum number of samples to collect.
        :return: Backend-specific TensorStatisticCollectorBase for the statistics calculation.
        """



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def get_input_masks(nx_node, nx_graph):
    """
    Return input masks for all inputs of nx_node.
    """
    input_edges = list(nx_graph.in_edges(nx_node["key"]))
    input_masks = [nx_graph.nodes[input_node]["output_mask"] for input_node, _ in input_edges]
    return input_masks


def identity_mask_propagation(nx_node, nx_graph):
    """
    Propagates input mask through nx_node.
    """
    input_masks = get_input_masks(nx_node, nx_graph)
    assert len(input_masks) == 1
    nx_node["input_masks"] = input_masks
    nx_node["output_mask"] = input_masks[0]



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from nncf.onnx.graph.metatypes import onnx_metatypes

QUANTIZE_AGNOSTIC_OPERATIONS = [
    onnx_metatypes.ONNXGlobalMaxPoolMetatype,
    onnx_metatypes.ONNXMaxPoolMetatype,
    onnx_metatypes.ONNXReduceMaxMetatype,
    onnx_metatypes.ONNXReshapeMetatype,
    onnx_metatypes.ONNXTransposeMetatype,
    onnx_metatypes.ONNXSqueezeMetatype,
    onnx_metatypes.ONNXUnsqueezeMetatype,
    onnx_metatypes.ONNXSplitMetatype,
    onnx_metatypes.ONNXTileMetatype,
    onnx_metatypes.ONNXCenterCropPadMetatype,
    onnx_metatypes.ONNXSliceMetatype,
    onnx_metatypes.ONNXPadMetatype,
    onnx_metatypes.ONNXGatherMetatype,
    onnx_metatypes.ONNXGatherNDMetatype,
    onnx_metatypes.ONNXGatherElementsMetatype,
    onnx_metatypes.ONNXDepthToSpaceMetatype,
    onnx_metatypes.ONNXSpaceToDepthMetatype,
    onnx_metatypes.ONNXScatterElementsMetatype,
    onnx_metatypes.ONNXScatterNDMetatype,
    onnx_metatypes.ONNXScatterMetatype,
    onnx_metatypes.ONNXCastLikeMetatype,
    onnx_metatypes.ONNXDropoutMetatype,
    onnx_metatypes.ONNXFlattenMetatype,
    onnx_metatypes.ONNXExpandMetatype,
    onnx_metatypes.ONNXIdentityMetatype,
    # ONNXReluMetatype is not considered to be QUANTIZATION_AGNOSTIC, because:
    # 1. Runtime doesn't provide performance benefits by quantizing the stand-alone RELU's (ticket: 59548)
    # 2. It's frequently better for the end accuracy to have quantizers set up after the RELU
    # so that the input distribution to the quantizer is non-negative
    # and we can therefore have better quantization resolution while preserving the original dynamic range
]


MATMUL_METATYPES = [onnx_metatypes.ONNXGemmMetatype, onnx_metatypes.ONNXMatMulMetatype]


INPUTS_QUANTIZABLE_OPERATIONS = [
    onnx_metatypes.ONNXConvolutionMetatype,
    onnx_metatypes.ONNXDepthwiseConvolutionMetatype,
    onnx_metatypes.ONNXConvolutionTransposeMetatype,
    *MATMUL_METATYPES,
    onnx_metatypes.ONNXAveragePoolMetatype,
    onnx_metatypes.ONNXGlobalAveragePoolMetatype,
    onnx_metatypes.ONNXAddLayerMetatype,
    onnx_metatypes.ONNXSubMetatype,
    onnx_metatypes.ONNXMulLayerMetatype,
    onnx_metatypes.ONNXBatchNormMetatype,
    onnx_metatypes.ONNXHardSigmoidMetatype,
    onnx_metatypes.ONNXResizeMetatype,
    onnx_metatypes.ONNXPowMetatype,
    onnx_metatypes.ONNXReciprocalMetatype,
    onnx_metatypes.ONNXMaximumMetatype,
    onnx_metatypes.ONNXMinimumMetatype,
]


CONSTANT_WEIGHT_LAYER_METATYPES = [
    onnx_metatypes.ONNXConvolutionMetatype,
    onnx_metatypes.ONNXDepthwiseConvolutionMetatype,
    onnx_metatypes.ONNXConvolutionTransposeMetatype,
    onnx_metatypes.ONNXEmbeddingMetatype,
]


LINEAR_OPERATIONS = [
    onnx_metatypes.ONNXConvolutionMetatype,
    onnx_metatypes.ONNXDepthwiseConvolutionMetatype,
    onnx_metatypes.ONNXConvolutionTransposeMetatype,
    onnx_metatypes.ONNXDeformableConvolutionMetatype,
    *MATMUL_METATYPES,
]


ATOMIC_ACTIVATIONS_OPERATIONS = [
    onnx_metatypes.ONNXReluMetatype,
    onnx_metatypes.ONNXLeakyReluMetatype,
    onnx_metatypes.ONNXThresholdedReluMetatype,
    onnx_metatypes.ONNXEluMetatype,
    onnx_metatypes.ONNXPReluMetatype,
    onnx_metatypes.ONNXSigmoidMetatype,
    onnx_metatypes.ONNXHardSigmoidMetatype,
    onnx_metatypes.ONNXHardSwishMetatype,
]


ARITHMETIC_OPERATIONS = [
    onnx_metatypes.ONNXAddLayerMetatype,
    onnx_metatypes.ONNXSubMetatype,
    onnx_metatypes.ONNXMulLayerMetatype,
    onnx_metatypes.ONNXDivLayerMetatype,
]


OPERATIONS_WITH_WEIGHTS = [
    *CONSTANT_WEIGHT_LAYER_METATYPES,
    *MATMUL_METATYPES,
]


BATCH_NORMALIZATION_OPERATIONS = [
    onnx_metatypes.ONNXBatchNormMetatype,
]


# Contains the operation metatypes for which bias can be applied.
OPERATIONS_WITH_BIAS_REDUCED = [
    onnx_metatypes.ONNXConvolutionMetatype,
    onnx_metatypes.ONNXGemmMetatype,
    # TODO: Need to add MatMul with the separate bias support (CVS-135433)
]

OPERATIONS_WITH_BIAS = [*OPERATIONS_WITH_BIAS_REDUCED, onnx_metatypes.ONNXDepthwiseConvolutionMetatype]


QUANTIZE_DEQUANTIZE_OPERATIONS = [
    onnx_metatypes.ONNXQuantizeLinearMetatype,
    onnx_metatypes.ONNXDequantizeLinearMetatype,
]

# These metatypes mix outputs for different samples into one axis.
# If reducers and aggregators collect statistics at the output of the following operations,
# assuming that 0-axis is batch axis, they get only 1 value instead of batch_size values.
# It could lead to inaccurate/incorrect statistics result.
OPERATIONS_OUTPUT_HAS_NO_BATCH_AXIS = [
    onnx_metatypes.ONNXROIAlignMetatype,
    onnx_metatypes.ONNXEmbeddingMetatype,
]



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import Counter
from copy import deepcopy
from typing import Dict, List, Set, Tuple, Union

import numpy as np
import onnx

import nncf
from nncf.common.graph.model_transformer import ModelTransformer
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.graph.transformations.layout import TransformationLayout
from nncf.onnx.graph.node_utils import get_input_edge
from nncf.onnx.graph.onnx_helper import get_children
from nncf.onnx.graph.onnx_helper import get_children_node_mapping
from nncf.onnx.graph.onnx_helper import get_edge_dtype
from nncf.onnx.graph.onnx_helper import get_edge_info_mapping
from nncf.onnx.graph.onnx_helper import get_name_to_node_map
from nncf.onnx.graph.onnx_helper import get_node_index
from nncf.onnx.graph.onnx_helper import get_tensor
from nncf.onnx.graph.transformations.commands import ONNXInitializerUpdateCommand
from nncf.onnx.graph.transformations.commands import ONNXModelExtractionCommand
from nncf.onnx.graph.transformations.commands import ONNXOutputInsertionCommand
from nncf.onnx.graph.transformations.commands import ONNXQDQNodeRemovingCommand
from nncf.onnx.graph.transformations.commands import ONNXQuantizerInsertionCommand


class ONNXModelTransformer(ModelTransformer):
    """
    Applies transformations upon ONNX model.
    ModelTransformer should be created once for a particular model,
    and be used to apply transformations to the provided model.
    """

    QUANTIZER_NAME_PREFIX = "QuantizeLinear_"
    DEQUANTIZER_NAME_PREFIX = "DequantizeLinear_"
    SCALE_TENSOR_NAME_PREFIX = "scale_"
    ZERO_POINT_NAME_PREFIX = "zero_point_"

    def __init__(self, model: onnx.ModelProto):
        infered_model = onnx.shape_inference.infer_shapes(model)
        super().__init__(infered_model)
        self.onnx_model_extractor = onnx.utils.Extractor(self._model)

    def _get_target_edge(
        self,
        port_id: int,
        node_name: str,
        transform_type: TargetType,
        node_mapping: Dict[str, onnx.NodeProto],
        input_edges_mapping: Dict[str, str],
    ) -> str:
        """
        Returns edge name corresponding to the node with a name equal to node_name, port_id and transform_type.

        :param port_id: Edge number of port.
        :param node_name: Node name.
        :param transform_type: Type of transformation.
        :param node_mapping: Mapping from a node name to the node.
        :param input_edges_mapping: Mapping between NNCF Input nodes and
        the following ONNX nodes and corresponding input port id.
        :return: Target edge name.
        """
        if transform_type in [TargetType.PRE_LAYER_OPERATION, TargetType.OPERATION_WITH_WEIGHTS]:
            return node_mapping[node_name].input[port_id]
        if node_name in input_edges_mapping:  # ADD INPUT NODE CASE
            return get_input_edge(node_name, input_edges_mapping, node_mapping)
        return node_mapping[node_name].output[port_id]

    def transform(self, transformation_layout: TransformationLayout) -> onnx.ModelProto:
        """
        Applies transformations to the model using an out-of-place approach.
        The transformations do not affect the original model, and a new model
        is returned with the transformations applied. If there are no transformations,
        returns a new instance of the original model.

        :param transformation_layout: Transformation commands.
        :return: The new instance of a model with applied transformations.
        """
        quantizer_insert_transformations = []
        output_insert_transformations = []
        initializer_update_transformations = []
        qdq_node_removing_transformations = []
        model_extraction_transformation = None
        transformations = transformation_layout.transformations
        # No transformation applied
        if not transformations:
            return deepcopy(self._model)
        for transformation in transformations:
            if isinstance(transformation, ONNXQuantizerInsertionCommand):
                quantizer_insert_transformations.append(transformation)
            elif isinstance(transformation, ONNXOutputInsertionCommand):
                output_insert_transformations.append(transformation)
            elif isinstance(transformation, ONNXModelExtractionCommand):
                model_extraction_transformation = transformation
            elif isinstance(transformation, ONNXQDQNodeRemovingCommand):
                qdq_node_removing_transformations.append(transformation)
            elif isinstance(transformation, ONNXInitializerUpdateCommand):
                initializer_update_transformations.append(transformation)
        # Inplace transformations, using deepcopy of model
        if quantizer_insert_transformations or initializer_update_transformations or qdq_node_removing_transformations:
            model = deepcopy(self._model)
            if quantizer_insert_transformations:
                model = self._apply_quantizer_insertion_transformations(model, quantizer_insert_transformations)
            if qdq_node_removing_transformations:
                model = self._apply_qdq_node_removing_transformations(model, qdq_node_removing_transformations)
            if initializer_update_transformations:
                model = self._apply_initializer_update_transformations(model, initializer_update_transformations)
        # Transformations that create new model
        if output_insert_transformations:
            model = self._apply_output_insertion_transformations(output_insert_transformations)
        if model_extraction_transformation:
            model = self._apply_model_extraction_transformation(model_extraction_transformation)
        return model

    def _apply_output_insertion_transformations(
        self, transformations: List[ONNXOutputInsertionCommand]
    ) -> onnx.ModelProto:
        """
        Returns a new model with extra outputs provided by transformations.

        :param transformations: ONNXOutputInsertionCommand transformations.
        :return: New model with inserted outputs.
        """
        model_outputs = set(output.name for output in self._model.graph.output)
        node_mapping = get_name_to_node_map(self._model)
        for transformation in transformations:
            port_id = transformation.target_point.port_id
            node_name = transformation.target_point.target_node_name
            transform_type = transformation.target_point.type
            input_edges_mapping = transformation.input_edges_mapping
            target_edge_name = self._get_target_edge(
                port_id, node_name, transform_type, node_mapping, input_edges_mapping
            )
            model_outputs.add(target_edge_name)

        return ONNXModelTransformer._insert_outputs(self._model, outputs=model_outputs)

    @staticmethod
    def _insert_outputs(model: onnx.ModelProto, outputs: Union[List[str], Set[str]]) -> onnx.ModelProto:
        """
        Creates a new model as a copy of provided model with additional outputs.

        :param model: Model of which copy will be created.
        :param outputs: Edge names to use as outputs.
        :return: New model with inserted outputs.
        """
        model_outputs = []
        edge_info_mapping = get_edge_info_mapping(model)
        for output in outputs:
            edge = edge_info_mapping[output]
            onnx_dtype = get_edge_dtype(edge)
            type_proto = onnx.helper.make_tensor_type_proto(onnx_dtype, shape=None)
            model_outputs.append(onnx.helper.make_value_info(name=output, type_proto=type_proto))

        graph = onnx.helper.make_graph(
            nodes=model.graph.node,
            name=model.graph.name,
            inputs=model.graph.input,
            outputs=model_outputs,
            initializer=model.graph.initializer,
            value_info=model.graph.value_info,
        )
        new_model = onnx.helper.make_model(
            graph,
            ir_version=model.ir_version,
            producer_name=model.producer_name,
            producer_version=model.producer_version,
            domain=model.domain,
            model_version=model.model_version,
            doc_string=model.doc_string,
        )
        if model.metadata_props:
            values = {p.key: p.value for p in model.metadata_props}
            onnx.helper.set_model_props(new_model, values)
        del new_model.opset_import[:]
        for oimp in model.opset_import:
            op_set = new_model.opset_import.add()
            op_set.domain = oimp.domain
            op_set.version = oimp.version
        return new_model

    def _apply_quantizer_insertion_transformations(
        self, model: onnx.ModelProto, transformations: List[ONNXQuantizerInsertionCommand]
    ) -> onnx.ModelProto:
        """
        Creates a new model as a deepcopy of provided model and inserts QuantizeLinear-DequantizeLinear nodes pair.

        :param model: Model to apply transformations.
        :param transformations: QuantizeLinear-DequantizeLinear nodes pair insertion transformation commands.
        :return: New model with inserted QuantizeLinear-DequantizeLinear nodes pairs.
        """
        self._added_target_edges = Counter()
        node_mapping = get_name_to_node_map(model)
        children_node_mapping = get_children_node_mapping(model)
        for transformation in transformations:
            model = self._insert_quantizer_dequantizer(model, transformation, node_mapping, children_node_mapping)
        return model

    def _get_quantize_dequantize_nodes(
        self, transformation: ONNXQuantizerInsertionCommand, target_edge_name: str
    ) -> Tuple[onnx.NodeProto, onnx.NodeProto]:
        """
        Returns QuantizeLinear-DequantizeLinear nodes pair, based on the transformation parameters and
        inserted onto edge with name target_edge_name.

        :param transformation: QuantizeLinear-DequantizeLinear insertion transformation,
        from which quantization axis is obtained.
        :param target_edge_name: Edge name on which QuantizeLinear-DequantizeLinear nodes pair should be placed.
        :return: QuantizeLinear-DequantizeLinear nodes pair.
        """
        axis = transformation.quantizer_parameters.axis

        cnt = self._added_target_edges[target_edge_name]

        input_target_edge = target_edge_name
        q_target_edge_name = target_edge_name + "_" + str(cnt)
        quantizer_name = ONNXModelTransformer.QUANTIZER_NAME_PREFIX + q_target_edge_name
        dequantizer_name = ONNXModelTransformer.DEQUANTIZER_NAME_PREFIX + q_target_edge_name
        scale_tensor_name = ONNXModelTransformer.SCALE_TENSOR_NAME_PREFIX + q_target_edge_name
        zero_point_tensor_name = ONNXModelTransformer.ZERO_POINT_NAME_PREFIX + q_target_edge_name

        quantizer = onnx.helper.make_node(
            name=quantizer_name,
            op_type="QuantizeLinear",
            inputs=[input_target_edge, scale_tensor_name, zero_point_tensor_name],
            outputs=["q_output_" + q_target_edge_name],
            axis=axis,
        )

        dequantizer = onnx.helper.make_node(
            name=dequantizer_name,
            op_type="DequantizeLinear",
            inputs=["q_output_" + q_target_edge_name, scale_tensor_name, zero_point_tensor_name],
            outputs=["dq_output_" + q_target_edge_name],
            axis=axis,
        )

        return quantizer, dequantizer

    @staticmethod
    def _get_scale_zero_point_tensors(
        transformation: ONNXQuantizerInsertionCommand, quantizer: onnx.NodeProto, dequantizer: onnx.NodeProto
    ) -> Tuple[onnx.TensorProto, onnx.TensorProto]:
        """
        Returns scale and zero point of QuantizeLinear-DequantizeLinear nodes pair.

        :param transformation: QuantizeLinear-DequantizeLinear insertion transformation,
        from which scale and zero point values are obtained.
        :param quantizer: QuantizeLinear node.
        :param dequantizer: DequantizeLinear node.
        :return: Scale and zero point tensors.
        """
        scale = transformation.quantizer_parameters.scale
        zero_point = transformation.quantizer_parameters.zero_point
        tensor_type = transformation.quantizer_parameters.tensor_type

        per_channel = scale.ndim > 0
        dims = scale.shape if per_channel else []
        onnx_scale = [scale.tolist()] if not per_channel else scale
        onnx_zero_point = [zero_point.tolist()] if not per_channel else zero_point
        if tensor_type == np.uint8:
            onnx_tensor_type = onnx.TensorProto.UINT8
        elif tensor_type == np.int8:
            onnx_tensor_type = onnx.TensorProto.INT8
        else:
            raise nncf.ValidationError(f"Incorrect tensor type - {tensor_type}.")
        assert quantizer.input[1] == dequantizer.input[1] and quantizer.input[2] == dequantizer.input[2]
        scale_tensor_name = quantizer.input[1]
        zero_point_tensor_name = quantizer.input[2]
        onnx_scale_tensor = onnx.helper.make_tensor(scale_tensor_name, onnx.TensorProto.FLOAT, dims, onnx_scale)
        onnx_zero_point_tensor = onnx.helper.make_tensor(
            zero_point_tensor_name, onnx_tensor_type, dims, onnx_zero_point
        )
        return onnx_scale_tensor, onnx_zero_point_tensor

    def _get_quantizer_dequantizer_edge_name(
        self, transformation: ONNXQuantizerInsertionCommand, node_mapping: Dict[str, onnx.NodeProto]
    ) -> str:
        """
        Returns an edge name on which QuantizeLinear-DequantizeLinear nodes pair has to be inserted.

        :param transformation: QuantizeLinear-DequantizeLinear insertion transformation.
        :param node_mapping: Mapping from a node name to the node.
        :return: Edge name to insert QuantizeLinear-DequantizeLinear nodes pair.
        """
        port_id = transformation.target_point.port_id
        node_name = transformation.target_point.target_node_name
        transform_type = transformation.target_point.type
        input_edges_mapping = transformation.input_edges_mapping
        target_edge_name = self._get_target_edge(port_id, node_name, transform_type, node_mapping, input_edges_mapping)
        self._added_target_edges[target_edge_name] += 1
        return target_edge_name

    def _insert_quantizer_dequantizer(
        self,
        model: onnx.ModelProto,
        transformation: ONNXQuantizerInsertionCommand,
        node_mapping: Dict[str, onnx.NodeProto],
        children_node_mapping: Dict[str, List[onnx.ValueInfoProto]],
    ) -> onnx.ModelProto:
        """
        Inserts QuantizeLinear-DequantizeLinear nodes pair.

        :param model: Model to insert new nodes.
        :param transformation: QuantizeLinear-DequantizeLinear insertion transformation.
        :param node_mapping: Mapping from node name to the node.
        :param children_node_mapping: Mapping from edge name to nodes which consume this edge as an input.
        :return: Updated model with inserted QuantizeLinear-DequantizeLinear pair.
        """
        target_edge_name = self._get_quantizer_dequantizer_edge_name(transformation, node_mapping)
        quantizer, dequantizer = self._get_quantize_dequantize_nodes(transformation, target_edge_name)
        onnx_scale_tensor, onnx_zero_point_tensor = ONNXModelTransformer._get_scale_zero_point_tensors(
            transformation, quantizer, dequantizer
        )

        # If several nodes on one edge
        input_nodes = []
        input_nodes.extend(children_node_mapping[target_edge_name])
        if not input_nodes:
            raise nncf.InternalError(
                f"Can not add the quantizer to the {target_edge_name} edge. This edge does not have end node."
            )

        if transformation.target_point.type == TargetType.PRE_LAYER_OPERATION:
            # If we need to change only target nodes input
            target_node = node_mapping[transformation.target_point.target_node_name]
            for i, inp in enumerate(target_node.input):
                if inp == target_edge_name:
                    target_node.input[i] = dequantizer.output[0]
        else:
            for node in input_nodes:
                for i, inp in enumerate(node.input):
                    if inp == target_edge_name:
                        node.input[i] = dequantizer.output[0]

        onnx_scale_value_info = onnx.helper.make_tensor_value_info(
            onnx_scale_tensor.name, onnx_scale_tensor.data_type, onnx_scale_tensor.dims
        )
        onnx_zero_point_info = onnx.helper.make_tensor_value_info(
            onnx_zero_point_tensor.name, onnx_zero_point_tensor.data_type, onnx_zero_point_tensor.dims
        )
        model.graph.initializer.extend([onnx_scale_tensor, onnx_zero_point_tensor])
        model.graph.value_info.extend([onnx_scale_value_info, onnx_zero_point_info])
        insert_index = get_node_index(model, input_nodes[0].name)
        model.graph.node.insert(insert_index, quantizer)
        model.graph.node.insert(insert_index + 1, dequantizer)
        return model

    def _apply_initializer_update_transformations(
        self, model: onnx.ModelProto, transformations: List[ONNXInitializerUpdateCommand]
    ) -> onnx.ModelProto:
        """
        Creates a copy of original model and applies bias correction transformations on the model.

        :param model: Model to apply transformations.
        :param transformations: Bias correction transformations.
        :return: Copy of original model with updated biases.
        """
        name_to_node_map = get_name_to_node_map(model)
        for transformation in transformations:
            node = name_to_node_map[transformation.target_point.target_node_name]
            initializer_name = node.input[transformation.target_point.port_id]
            initializer = get_tensor(model, initializer_name)

            new_tensor = onnx.numpy_helper.from_array(transformation.new_value, initializer_name)
            initializer.CopyFrom(new_tensor)
        return model

    def _apply_model_extraction_transformation(self, transformation: ONNXModelExtractionCommand) -> onnx.ModelProto:
        """
        Returns a new model that is a sub-model from the original between provided inputs and outputs.

        :param transformation: Model extraction transformation.
        :return: Extracted sub-model.
        """
        input_tensor_names = []
        node_mapping = get_name_to_node_map(self._model)
        for input_name, input_port_id in transformation.input_ids:
            input_onnx_node = node_mapping[input_name]
            input_tensor_names.append(input_onnx_node.input[input_port_id])

        output_tensor_names = []
        for output_name, output_port_id in transformation.output_ids:
            output_onnx_node = node_mapping[output_name]
            output_tensor_names.append(output_onnx_node.output[output_port_id])

        if not output_tensor_names:
            output_tensor_names = [n.name for n in self._model.graph.output]

        return self.onnx_model_extractor.extract_model(input_tensor_names, output_tensor_names)

    def _apply_qdq_node_removing_transformations(
        self, model: onnx.ModelProto, transformations: List[ONNXQDQNodeRemovingCommand]
    ) -> onnx.ModelProto:
        """
        Returns a copy of original model with removed nodes.

        :param model: Model to apply transformations.
        :param transformations: Nodes removing transformations.
        :return: Model with removed nodes.
        """
        name_to_node_map = get_name_to_node_map(model)
        children_node_mapping = get_children_node_mapping(model)
        # We combine quantize and dequantize nodes into pairs here because it
        # does not make sense to remove only the quantize node or the dequantize
        # node. They should be removed together.
        was_processed = {t.target_point.target_node_name: False for t in transformations}
        quantize_dequantize_pairs = []
        for node_name in was_processed:
            if was_processed[node_name]:
                continue
            quantize_node_proto = name_to_node_map[node_name]
            if quantize_node_proto.op_type != "QuantizeLinear":
                continue
            # `quantize_node_proto` has only one child, which is the dequantize node.
            dequantize_node_proto = next(iter(get_children(quantize_node_proto, children_node_mapping)))
            assert dequantize_node_proto.op_type == "DequantizeLinear"

            quantize_dequantize_pairs.append((quantize_node_proto, dequantize_node_proto))
            was_processed[quantize_node_proto.name] = True
            was_processed[dequantize_node_proto.name] = True

        if not all(was_processed.values()):
            raise RuntimeError("Invalid transformation commands.")

        initializers = {i.name: i for i in model.graph.initializer}
        value_infos = {i.name: i for i in model.graph.value_info}

        for quantize_node_proto, dequantize_node_proto in quantize_dequantize_pairs:
            # Unlink Q-DQ subgraph from graph
            children = get_children(dequantize_node_proto, children_node_mapping)
            for child in children:
                for port_id, input_name in enumerate(child.input):
                    if input_name == dequantize_node_proto.output[0]:
                        child.input[port_id] = quantize_node_proto.input[0]
            # QuantizeLinear and DequantizeLinear nodes have common initializers in ports 1 and 2.
            for i in [1, 2]:
                model.graph.initializer.remove(initializers[quantize_node_proto.input[i]])
                model.graph.value_info.remove(value_infos[quantize_node_proto.input[i]])

            for node_proto in [quantize_node_proto, dequantize_node_proto]:
                model.graph.value_info.remove(value_infos[node_proto.output[0]])

            model.graph.node.remove(quantize_node_proto)
            model.graph.node.remove(dequantize_node_proto)

        return model



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import random
from collections import OrderedDict
from contextlib import contextmanager
from typing import Any, Dict, Generator, List

import numpy as np
import torch
from torch import distributed as dist
from torch import nn
from torch.nn import Module

import nncf
from nncf.common.compression import BaseCompressionAlgorithmController as BaseController
from nncf.common.deprecation import warning_deprecated
from nncf.common.graph import NNCFNodeName
from nncf.common.logging import nncf_logger
from nncf.common.scopes import matches_any
from nncf.torch.dynamic_graph.scope import Scope
from nncf.torch.dynamic_graph.scope import ScopeElement
from nncf.torch.dynamic_graph.trace_tensor import TracedTensorMixin
from nncf.torch.layer_utils import _NNCFModuleMixin
from nncf.torch.structures import ExecutionParameters


def get_node_name(module, module_name, prefix):
    return "{prefix}/{cls}[{name}]".format(prefix=prefix, cls=module.__class__.__name__, name=module_name)


def get_all_modules(model, prefix=None):
    found = OrderedDict()
    if prefix is None:
        prefix = model.__class__.__name__
    for name, module in model.named_children():
        full_node_name = get_node_name(module, name, prefix)
        found[full_node_name] = module
        sub_found = get_all_modules(module, prefix=full_node_name)
        if sub_found:
            found.update(sub_found)
    return found


def get_all_modules_by_type(
    model, module_types=None, current_scope=None, ignored_scopes=None, target_scopes=None, memo=None
) -> Dict[Scope, Module]:
    if memo is None:
        memo = set()
    if isinstance(module_types, str):
        module_types = [module_types]
    found = OrderedDict()

    if current_scope is None:
        current_scope = Scope()
        current_scope.push(ScopeElement(model.__class__.__name__))
    for name, module in model.named_children():
        if id(module) in memo:
            continue
        memo.add(id(module))
        child_scope_element = ScopeElement(module.__class__.__name__, name)
        child_scope = current_scope.copy()
        child_scope.push(child_scope_element)

        if matches_any(str(child_scope), ignored_scopes):
            continue

        if target_scopes is None or matches_any(str(child_scope), target_scopes):
            if module_types is None or module_types.count(str(type(module).__name__)) != 0:
                found[child_scope] = module
            sub_found = get_all_modules_by_type(
                module,
                module_types,
                current_scope=child_scope,
                ignored_scopes=ignored_scopes,
                target_scopes=target_scopes,
                memo=memo,
            )
            if sub_found:
                found.update(sub_found)
    return found


def get_state_dict_names_with_modules(
    model: torch.nn.Module, str_types: List[str] = None, prefix=""
) -> Dict[str, torch.nn.Module]:
    found = OrderedDict()
    for name, module in model.named_children():
        full_node_name = "{}{}".format(prefix, name)
        if str_types is not None and type(module).__name__ in str_types:
            found[full_node_name] = module
        sub_found = get_state_dict_names_with_modules(module, str_types, prefix=full_node_name + ".")
        if sub_found:
            found.update(sub_found)
    return found


def get_filters_num(module):
    if isinstance(module, _NNCFModuleMixin):
        return module.weight.size(module.target_weight_dim_for_compression)
    return module.weight.size(0)


def manual_seed(seed):
    np.random.seed(seed)
    random.seed(seed)
    torch.manual_seed(seed)


def is_tracing_state():
    return torch._C._get_tracing_state() is not None


class no_jit_trace:
    def __enter__(self):
        self.state = torch._C._get_tracing_state()
        torch._C._set_tracing_state(None)

    def __exit__(self, *args):
        torch._C._set_tracing_state(self.state)
        self.state = None


def fp32_accum_wrapper(func):
    def wrapper(tensor_to_sum, ret_tensor):
        half = tensor_to_sum.dtype == np.float16
        if half:
            tensor_to_sum = tensor_to_sum.astype(np.float32)
        retval = func(tensor_to_sum, ret_tensor)
        if half:
            retval = retval.astype(np.float16)
        return retval

    return wrapper


@fp32_accum_wrapper
def sum_like(tensor_to_sum, ref_tensor):
    """Warning: may modify tensor_to_sum"""
    if ref_tensor.size == 1:
        return tensor_to_sum.sum()

    for dim, size in enumerate(ref_tensor.shape):
        if size == 1:
            if isinstance(tensor_to_sum, np.ndarray):
                tensor_to_sum = tensor_to_sum.sum(dim, keepdims=True)
            else:
                tensor_to_sum = tensor_to_sum.sum(dim, keepdim=True)
    return tensor_to_sum


def get_flat_tensor_contents_string(input_tensor):
    retval = "["
    for idx, el in enumerate(input_tensor.view(-1)):
        if idx >= 10:
            retval += "... (first 10/{} elements shown only) ".format(len(input_tensor.view(-1)))
            break
        retval += "{:.4f}, ".format(el.item())
    retval += "]"
    return retval


def is_dist_avail_and_initialized():
    if not dist.is_available():
        return False
    if not dist.is_initialized():
        return False
    return True


def get_rank():
    if not is_dist_avail_and_initialized():
        return 0
    return dist.get_rank()


def is_main_process():
    return get_rank() == 0


def get_world_size():
    if not is_dist_avail_and_initialized():
        return 1
    return dist.get_world_size()


def safe_thread_call(main_call_fn, after_barrier_call_fn=None):
    result = None
    if is_dist_avail_and_initialized():
        if is_main_process():
            result = main_call_fn()
        dist.barrier()
        if not is_main_process():
            result = after_barrier_call_fn() if after_barrier_call_fn else main_call_fn()
    else:
        result = main_call_fn()
    return result


def is_tensor(obj):
    return isinstance(obj, torch.Tensor)


def is_traced_tensor(obj):
    return isinstance(obj, TracedTensorMixin)


class _ModuleState:
    def __init__(self, base_module: Module = None):
        self._training_state: Dict[str, bool] = {}
        self._requires_grad_state: Dict[str, bool] = {}
        if base_module is not None:
            for module_name, module in base_module.named_modules():
                self.training_state[module_name] = module.training

            for param_name, param in base_module.named_parameters():
                self.requires_grad_state[param_name] = param.requires_grad

    @property
    def training_state(self) -> Dict[str, bool]:
        return self._training_state

    @property
    def requires_grad_state(self) -> Dict[str, bool]:
        return self._requires_grad_state


def save_module_state(module: Module) -> _ModuleState:
    return _ModuleState(module)


def load_module_state(base_module: Module, state: _ModuleState, strict=False) -> None:
    for name, module in base_module.named_modules():
        try:
            module.train(state.training_state[name])
        except KeyError as e:
            # KeyError could happen if the modules name were changed during forward
            # (e.g. LSTM block in NNCF examples)
            msg = f"Could not find a module to restore state: {name}"
            nncf_logger.debug(msg)
            if strict:
                raise nncf.InternalError(msg) from e

    for name, param in base_module.named_parameters():
        param.requires_grad = state.requires_grad_state[name]


@contextmanager
def training_mode_switcher(model: Module, is_training: bool = True):
    saved_state = save_module_state(model)
    model.train(is_training)
    try:
        yield
    finally:
        load_module_state(model, saved_state)


def compute_FLOPs_hook(module, input_, output, dict_to_save, module_node_name: NNCFNodeName):
    # WARNING: numpy should be explicitly given np.int64 as dtype, since default integer type on Win is np.int32
    if isinstance(
        module, (nn.Conv1d, nn.ConvTranspose1d, nn.Conv2d, nn.ConvTranspose2d, nn.Conv3d, nn.ConvTranspose3d)
    ):
        ks = module.weight.data.shape
        mac_count = np.prod(ks, dtype=np.int64) * np.prod(output.shape[2:], dtype=np.int64)
    elif isinstance(module, nn.Linear):
        if len(input_[0].shape) == 1:
            # In some test cases input tensor could have dimension [N]
            mac_count = input_[0].shape[0] * output.shape[-1]
        else:
            mac_count = np.prod(input_[0].shape[1:], dtype=np.int64) * output.shape[-1]
    else:
        return
    dict_to_save[module_node_name] = 2 * mac_count


def add_domain(name_operator: str) -> str:
    from nncf.torch.compression_method_api import DOMAIN_CUSTOM_OPS_NAME

    return DOMAIN_CUSTOM_OPS_NAME + "::" + name_operator


def default_distributed_wrapper(model: nn.Module, execution_parameters: ExecutionParameters):
    """
    Wrapping model for distributed training with DataParallel or DistributedDataParallel depending on execution mode
    chosen by user.
    :param execution_parameters: structure with necessary execution parameters
    :param model: model to wrap  in accordance with execution mode chosen by user
    :return: wrapped model
    """
    if not execution_parameters or execution_parameters.cpu_only:
        # If execution params is not set or in cpu_only mode model can't be optimized by parallelization
        return model

    current_gpu = execution_parameters.current_gpu
    if not is_dist_avail_and_initialized():
        if current_gpu is not None:
            # ExecutionMode.SINGLE_GPU
            torch.cuda.set_device(current_gpu)
        else:
            # ExecutionMode.GPU_DATAPARALLEL
            model = torch.nn.DataParallel(model)
    else:
        if current_gpu is None:
            # ExecutionMode.DISTRIBUTED
            model = torch.nn.parallel.DistributedDataParallel(model)
        else:
            # ExecutionMode.MULTIPROCESSING_DISTRIBUTED
            torch.cuda.set_device(current_gpu)
            model = torch.nn.parallel.distributed.DistributedDataParallel(model, device_ids=[current_gpu])
    return model


def default_distributed_unwrapper(model: nn.Module):
    """
    Unwrapping model prepared from distributed training in Pytorch
    (and wrapped with Dataparallel or DistributedDataParallel).
    :param model: model to unwrap.
    :return: model without parallelization
    """
    if isinstance(model, (torch.nn.parallel.DataParallel, torch.nn.parallel.DistributedDataParallel)):
        return model.module
    return model


def rename_legacy_names_in_state_dict(
    state_dict_to_load: Dict[str, Any], legacy_names: List[str], legacy_name: str, new_name: str
):
    for name in legacy_names:
        tensor = state_dict_to_load.pop(name)
        new_key = name.replace(legacy_name, new_name) if new_name not in name else name
        state_dict_to_load[new_key] = tensor

    if legacy_names:
        warning_deprecated(
            "Legacy Batch Norm layer names was detected in checkpoint model state dict."
            " All occurrences of `{}` in nodes names was replaced by `{}`".format(legacy_name, new_name)
        )


LEGACY_VS_NEW_BN_MAP = {
    "BatchNorm1d": "NNCFBatchNorm1d",
    "BatchNorm2d": "NNCFBatchNorm2d",
    "BatchNorm3d": "NNCFBatchNorm3d",
    "NNCFBatchNorm": "NNCFBatchNorm2d",
    "ConvBNActivation": "Conv2dNormActivation",
}


def maybe_convert_legacy_names_in_model_state(state_dict_to_load: Dict[str, Any]) -> None:
    """
    Convert legacy layer names in compressed model state dict in case such names exist.

    :param state_dict_to_load: State dict to convert.
    """
    legacy_names = LEGACY_VS_NEW_BN_MAP.keys()
    matched_legacy_names = {name: [] for name in legacy_names}
    for name_in_state_dict in state_dict_to_load:
        matched = filter(lambda x: x in name_in_state_dict, legacy_names)
        for legacy_name in matched:
            matched_legacy_names[legacy_name].append(name_in_state_dict)
    for old_name, new_name in LEGACY_VS_NEW_BN_MAP.items():
        rename_legacy_names_in_state_dict(state_dict_to_load, matched_legacy_names[old_name], old_name, new_name)


def maybe_convert_legacy_names_in_compress_state(compression_state: Dict[str, Any]) -> None:
    """
    Convert legacy layer names in compression state in case such names exist.

    :param compression_state: Compression state to convert.
    """
    if not compression_state or BaseController.BUILDER_STATE not in compression_state:
        return

    controller_state = compression_state[BaseController.BUILDER_STATE]
    if not controller_state or "quantization" not in controller_state:
        return

    from nncf.torch.quantization.algo import QUANTIZER_BUILDER_STATE_VERSION_SAVE_NAME

    if not controller_state["quantization"].get(QUANTIZER_BUILDER_STATE_VERSION_SAVE_NAME):
        qips = controller_state["quantization"]["quantizer_setup"]["quantization_points"]

        detected_legacy_names = {
            "BatchNorm1d": False,
            "BatchNorm2d": False,
            "BatchNorm3d": False,
            "NNCFBatchNorm": False,
        }

        for point in qips.values():
            name = point["qip"]["target_node_name"]
            for old_name, new_name in LEGACY_VS_NEW_BN_MAP.items():
                if old_name in name and new_name not in name:
                    detected_legacy_names[old_name] = True
                    point["qip"]["target_node_name"] = name.replace(old_name, new_name)
                    break

        for old_name, was_detected in detected_legacy_names.items():
            if was_detected:
                new_name = LEGACY_VS_NEW_BN_MAP[old_name]
                warning_deprecated(
                    "Legacy Batch Norm layer names was detected in quantization setup target"
                    " point names. All occurrences of `{}` in nodes names was replaced by"
                    " `{}`".format(old_name, new_name)
                )


def get_model_device(model: torch.nn.Module) -> torch.device:
    """
    Get the device on which the first model parameters reside.

    :param model: The PyTorch model.
    :return: The device where the first model parameter reside.
        Default cpu if the model has no parameters.
    """

    try:
        device = next(model.parameters()).device
    except StopIteration:
        # The model had no parameters at all, doesn't matter which device to choose
        device = torch.device("cpu")
    return device


def get_all_model_devices_generator(model: torch.nn.Module) -> Generator[torch.device, None, None]:
    for p in model.parameters():
        yield p.device


def is_multidevice(model: torch.nn.Module) -> bool:
    """
    Checks if the model's parameters are distributed across multiple devices.

    :param model: The PyTorch model.
    :return: True if the parameters reside on multiple devices, False otherwise.
        Default False if the models has no parameters
    """

    device_generator = get_all_model_devices_generator(model)
    try:
        curr_device = next(device_generator)
    except StopIteration:  # no parameters
        return False

    for d in device_generator:
        if d != curr_device:
            return True
    return False


def get_model_dtype(model: torch.nn.Module) -> torch.dtype:
    """
    Get the datatype of the first model parameter.

    :param model: The PyTorch model.
    :return: The datatype of the first model parameter.
        Default to torch.float32 if the model has no parameters.
    """

    try:
        dtype = next(model.parameters()).dtype
    except StopIteration:
        # The model had no parameters at all, assume FP32
        dtype = torch.float32
    return dtype



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from nncf.common.graph.patterns import merge_two_types_of_operations
from nncf.tensorflow.graph.metatypes.common import ELEMENTWISE_LAYER_METATYPES
from nncf.tensorflow.graph.metatypes.common import GENERAL_CONV_LAYER_METATYPES
from nncf.tensorflow.graph.metatypes.common import LAYER_METATYPES_AGNOSTIC_TO_DATA_PRECISION_WITH_ONE_INPUT
from nncf.tensorflow.graph.metatypes.common import LINEAR_LAYER_METATYPES

LINEAR_OPERATIONS = {
    "type": list(
        {
            *{layer_name for m in GENERAL_CONV_LAYER_METATYPES for layer_name in m.get_all_aliases()},
            *{layer_name for m in LINEAR_LAYER_METATYPES for layer_name in m.get_all_aliases()},
        }
    ),
    "label": "LINEAR",
}

ELEMENTWISE_OPERATIONS = {
    "type": list(set(layer_name for m in ELEMENTWISE_LAYER_METATYPES for layer_name in m.get_all_aliases())),
    "label": "ELEMENTWISE",
}

QUANTIZATION_AGNOSTIC_OPERATIONS = {
    "type": list(
        set(
            layer_name
            for m in LAYER_METATYPES_AGNOSTIC_TO_DATA_PRECISION_WITH_ONE_INPUT
            for layer_name in m.get_all_aliases()
        )
    ),
    "label": "QUANTIZATION_AGNOSTIC",
}

BATCH_NORMALIZATION_OPERATIONS = {
    "type": ["BatchNormalization", "SyncBatchNormalization", "FusedBatchNormV3"],
    "label": "BATCH_NORMALIZATION",
}

KERAS_ACTIVATIONS_OPERATIONS = {
    "type": ["ReLU", "ThresholdedReLU", "ELU", "PReLU", "LeakyReLU", "Activation"],
    "label": "KERAS_ACTIVATIONS",
}


TF_ACTIVATIONS_OPERATIONS = {
    "type": [
        "Relu",
        "nn.relu",
        "Elu",
        "LeakyRelu",
        "Relu6",
        "Selu",
        "Sigmoid",
        "Tanh",
    ],
    "label": "TF_ACTIVATIONS",
}

ATOMIC_ACTIVATIONS_OPERATIONS = merge_two_types_of_operations(
    KERAS_ACTIVATIONS_OPERATIONS, TF_ACTIVATIONS_OPERATIONS, "ATOMIC_ACTIVATIONS"
)

POOLING_OPERATIONS = {
    "type": [
        "AveragePooling2D",
        "AveragePooling3D",
        "GlobalAveragePooling2D",
        "GlobalAveragePooling3D",
        "AvgPool",
        "AvgPool3D",
        "Mean",
    ],
    "label": "POOLING",
}

SINGLE_OPS = merge_two_types_of_operations(
    POOLING_OPERATIONS, {"type": ["Average", "LayerNormalization", "UpSampling2D"]}, label="SINGLE_OPS"
)

ARITHMETIC_OPERATIONS = {"type": ["__iadd__", "__add__", "__mul__", "__rmul__"], "label": "ARITHMETIC"}



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from collections import defaultdict
from enum import Enum
from enum import auto
from functools import reduce
from operator import not_
from typing import Dict, List, Optional, Tuple, Type, Union

from nncf.common.graph.graph import NNCFGraph
from nncf.common.graph.graph import NNCFNode
from nncf.common.graph.layer_attributes import GetItemLayerAttributes
from nncf.common.graph.layer_attributes import GroupNormLayerAttributes
from nncf.common.graph.layer_attributes import MultipleOutputLayerAttributes
from nncf.common.graph.layer_attributes import PermuteLayerAttributes
from nncf.common.graph.layer_attributes import ReshapeLayerAttributes
from nncf.common.graph.layer_attributes import TransposeLayerAttributes
from nncf.common.logging import nncf_logger
from nncf.common.pruning.tensor_processor import NNCFPruningBaseTensorProcessor
from nncf.common.pruning.utils import get_input_masks
from nncf.common.pruning.utils import identity_mask_propagation
from nncf.experimental.common.pruning.nodes_grouping import PropagationMask
from nncf.experimental.common.pruning.nodes_grouping import PruningBlock
from nncf.experimental.common.pruning.propagation_data import ConsumerInfo
from nncf.experimental.common.pruning.propagation_data import PropagationGroup


class BasePruningOp:
    """
    Determines meta operations which aggregate operations having common
    properties of interaction with pruning masks
    """

    subtypes = []
    additional_types = []

    @classmethod
    def mask_propagation(
        cls, node: NNCFNode, graph: NNCFGraph, tensor_processor: Type[NNCFPruningBaseTensorProcessor]
    ) -> None:
        """
        Propagates the pruning mask through a node using pruning masks of all inputs and the current node (if any).

        :param node: The graph node to propagate mask through it
        :param graph: The model graph to prune
        :param tensor_processor: Interface with tensor processing methods
        """
        raise NotImplementedError

    @classmethod
    def get_all_op_aliases(cls) -> List[str]:
        """
        :return: list of all aliases of types in metatype
        """
        op_types = []
        for subtype in cls.subtypes:
            op_types.extend(subtype.get_all_aliases())
        op_types = list(set(op_types)) + cls.additional_types
        return op_types

    @classmethod
    def invalidate_masks(cls, input_masks: List[Union[None, PropagationMask]]) -> None:
        """
        Safely invalidates groups for the list of Nones and propagation masks.
        Usually such output comes from the `get_input_masks` helper.

        :param input_masks: list of input masks or N
        """
        for mask in input_masks:
            if mask:
                mask.invalidate_groups()


class InputPruningOp(BasePruningOp):
    @classmethod
    def mask_propagation(
        cls, node: NNCFNode, graph: NNCFGraph, tensor_processor: Type[NNCFPruningBaseTensorProcessor]
    ) -> None:
        node.attributes["output_mask"] = None


class OutputPruningOp(BasePruningOp):
    @classmethod
    def mask_propagation(
        cls, node: NNCFNode, graph: NNCFGraph, tensor_processor: Type[NNCFPruningBaseTensorProcessor]
    ) -> None:
        node.attributes["output_mask"] = None
        input_masks = get_input_masks(node, graph)
        cls.invalidate_masks(input_masks)


class IdentityMaskForwardPruningOp(BasePruningOp):
    @classmethod
    def mask_propagation(
        cls, node: NNCFNode, graph: NNCFGraph, tensor_processor: Type[NNCFPruningBaseTensorProcessor]
    ) -> None:
        identity_mask_propagation(node, graph)


class LinearPruningOp(BasePruningOp):
    @classmethod
    def mask_propagation(
        cls, node: NNCFNode, graph: NNCFGraph, tensor_processor: Type[NNCFPruningBaseTensorProcessor]
    ) -> None:
        input_masks = get_input_masks(node, graph)
        assert len(input_masks) in [1, 2]
        is_input_mask_empty_map = list(map(not_, input_masks))
        output_mask = node.attributes.get("output_mask", None)
        input_tensors_shapes = [x.tensor_shape for x in graph.get_input_edges(node)]
        node_id = node.node_id
        if all(is_input_mask_empty_map):
            # It's acceptable to have empty PropagationMask on all inputs.
            # It means there's no pruning on this way to node. But the linear op can produce pruning mask,
            # need to propagate it, if it exists.
            pass
        elif any(is_input_mask_empty_map):
            # It's NOT acceptable to empty PropagationMask on one of the input and have non-empty
            # PropagationMask on another input at once. The mask is invalidated.
            cls.invalidate_masks(input_masks)
            output_mask = None
        elif len(input_masks) == 1:
            output_mask = cls._handle_single_input(input_masks[0], input_tensors_shapes, node_id, output_mask)
        elif len(input_masks) == 2:
            output_mask = cls._handle_two_inputs(input_masks, input_tensors_shapes, node_id)

        node.attributes["output_mask"] = output_mask

    @staticmethod
    def _handle_single_input(
        input_mask: PropagationMask,
        input_tensors_shapes: List[List[int]],
        node_id: int,
        output_mask: Optional[PropagationMask] = None,
    ) -> Optional[PropagationMask]:
        # Linear module with built-in weights. Assume single input for activations.
        input_shape_len = len(input_tensors_shapes[0])
        # Propagate further all except the last one, which is closing
        for dim, groups in input_mask.dim_groups_map.items():
            if dim == input_shape_len - 1:
                for group in groups:
                    consumer = ConsumerInfo(node_id=node_id, pruning_dimension=1)
                    group.add_consumers({consumer})
            else:
                if output_mask is None:
                    output_mask = PropagationMask()
                output_mask.dim_groups_map[dim] = groups
        return output_mask

    @staticmethod
    def _handle_two_inputs(
        input_masks: List[PropagationMask], input_tensors_shapes: List[List[int]], node_id: int
    ) -> PropagationMask:
        # Matmul operator with 2 inputs: one - for activation, second - for weights.
        assert len(input_tensors_shapes[0]) == len(input_tensors_shapes[1])
        input_shape_len = len(input_tensors_shapes[0])
        # Join consumed masks
        left_dim_groups, right_dim_groups = [input_masks[i].dim_groups_map for i in range(2)]

        def _both_dim_blocks_exist(left_idx, right_idx):
            if left_idx in left_dim_groups or right_idx in right_dim_groups:
                assert left_idx in left_dim_groups and right_idx in right_dim_groups
                return True
            return False

        output_mask = PropagationMask()
        # Propagating batch dims
        for dim in range(input_shape_len - 2):
            if _both_dim_blocks_exist(dim, dim):
                left = left_dim_groups[dim]
                right = right_dim_groups[dim]
                assert len(left) == 1 and len(right) == 1, "multiple groups are not supported"
                output_mask.dim_groups_map[dim] = [PropagationGroup.join_groups(left[0], right[0])]

        # Propagating left rows / right cols
        for idx, dim in enumerate(range(input_shape_len - 2, input_shape_len)):
            if dim in input_masks[idx].dim_groups_map:
                output_mask.dim_groups_map[dim] = input_masks[idx].dim_groups_map[dim]

        # Close branch
        if _both_dim_blocks_exist(input_shape_len - 1, input_shape_len - 2):
            left = left_dim_groups[input_shape_len - 1]
            right = right_dim_groups[input_shape_len - 2]
            assert len(left) == 1 and len(right) == 1, "multiple groups are not supported"
            group = PropagationGroup.join_groups(left[0], right[0])
            consumer = ConsumerInfo(node_id=node_id, pruning_dimension=None)
            group.add_consumers({consumer})
        return output_mask


class BatchNormPruningOp(BasePruningOp):
    @classmethod
    def mask_propagation(
        cls, node: NNCFNode, graph: NNCFGraph, tensor_processor: Type[NNCFPruningBaseTensorProcessor]
    ) -> None:
        identity_mask_propagation(node, graph)


class GroupNormPruningOp(BasePruningOp):
    @classmethod
    def accept_pruned_input(cls, node: NNCFNode) -> bool:
        # For Instance Normalization
        return (
            isinstance(node.layer_attributes, GroupNormLayerAttributes)
            and node.layer_attributes.num_groups == node.layer_attributes.num_channels
        )

    @classmethod
    def mask_propagation(
        cls, node: NNCFNode, graph: NNCFGraph, tensor_processor: Type[NNCFPruningBaseTensorProcessor]
    ) -> None:
        if cls.accept_pruned_input(node):
            identity_mask_propagation(node, graph)
        else:
            node.attributes["output_mask"] = None


class ElementwisePruningOp(BasePruningOp):
    @classmethod
    def mask_propagation(
        cls, node: NNCFNode, graph: NNCFGraph, tensor_processor: Type[NNCFPruningBaseTensorProcessor]
    ) -> None:
        input_masks = get_input_masks(node, graph)
        input_shapes = [x.tensor_shape for x in graph.get_input_edges(node)]
        node.attributes["output_mask"] = cls._get_output_mask(input_masks, input_shapes)

    @classmethod
    def _get_output_mask(
        cls, input_masks: List[Optional[PropagationMask]], input_shapes: List[Tuple[int, ...]]
    ) -> Optional[PropagationMask]:
        if not input_masks:
            return None
        output_mask = None
        if len(input_masks) == 1:
            nncf_logger.warning(
                "ElementWise with a single input is not properly supported. "
                "The second input might be a constant without node in the graph. "
                "The constant should be in the graph or in the node attributes. "
                "It's also should be pruned in accordance with an input mask. "
                "node_name={node.node_name}"
            )
            output_mask = input_masks[0]
        elif any(m is None for m in input_masks) and any(m is not None for m in input_masks):
            # In case of one from input_masks is None
            output_mask = cls._propagate_single_mask(input_masks, input_shapes)
            if output_mask is None:
                cls.invalidate_masks(input_masks)
        elif any(not m for m in input_masks):
            # Need non-empty masks on all branches in order to properly propagate pruning mask,
            # otherwise - invalidate masks
            cls.invalidate_masks(input_masks)
        elif all(m is not None for m in input_masks):
            # Each branch/mask should have a single group along the same dimension. These groups are joined, all others
            # are invalidated.
            output_mask = PropagationMask()
            first_mask = input_masks[0]
            dims_to_invalidate = []
            for dim in first_mask.dim_groups_map:
                is_dim_in_masks = all(dim in m.dim_groups_map for m in input_masks[1:])
                if is_dim_in_masks:
                    is_one_group_per_dim = all(len(m.dim_groups_map[dim]) == 1 for m in input_masks)
                    if is_one_group_per_dim:
                        groups_to_join = [m.dim_groups_map[dim][0] for m in input_masks]
                        output_mask.dim_groups_map[dim] = [PropagationGroup.join_groups(*groups_to_join)]
                    else:
                        dims_to_invalidate.append(dim)
                else:
                    dims_to_invalidate.append(dim)

            for dim in dims_to_invalidate:
                for m in input_masks:
                    if dim in m.dim_groups_map:
                        for group in m.dim_groups_map[dim]:
                            group.invalidate()
        return output_mask

    @classmethod
    def _propagate_single_mask(
        cls, input_masks: List[Optional[PropagationMask]], input_shapes: List[Tuple[int, ...]]
    ) -> Optional[PropagationMask]:
        """
        Attempts to propagate a mask in case of one input mask is None.

        :param input_masks: List of propagation masks for each input of the element-wise operation
        :param input_shapes: List of tensor shapes for each input.
        :return: An instance of PropagationMask or None.
        """
        if cls._are_broadcast_dims_in_both_shapes(input_shapes):
            return None

        none_mask_ind = input_masks.index(None)
        mask_ind = 0 if none_mask_ind else 1

        dims_diff = len(input_shapes[mask_ind]) - len(input_shapes[none_mask_ind])
        padded_none_mask_shape = (1,) * dims_diff + input_shapes[none_mask_ind]

        dims_shift = min(dims_diff, 0)
        for dim in input_masks[mask_ind].dim_groups_map:
            if padded_none_mask_shape[dim - dims_shift] != 1:
                return None

        output_mask = PropagationMask()
        for dim, groups in input_masks[mask_ind].dim_groups_map.items():
            output_mask.dim_groups_map[dim - dims_shift] = groups

        return output_mask

    @staticmethod
    def _are_broadcast_dims_in_both_shapes(shapes) -> bool:
        """
        Propagation mask is not supported if both shapes will broadcasting by elementwise operation.
        True, if both shapes have broadcasted dimensions, otherwise False.

        Example:
            (1, 10), (10, 1) -> True
            (1,10), (10,) -> False

        :param shapes: Shapes of tensors.
        :return: True, if both shapes have broadcasted dimensions, otherwise False.
        """
        shape_a = shapes[0]
        shape_b = shapes[1]
        shape_a_size_diff = len(shape_a) - len(shape_b)
        shape_b_size_diff = -shape_a_size_diff
        broadcasted_dims_1 = set()
        broadcasted_dims_2 = set()

        for i in range(len(shape_a)):
            shifted_elem = i + shape_b_size_diff
            if shifted_elem >= 0 and shape_a[i] == 1 and shape_b[shifted_elem] != 1:
                broadcasted_dims_1.add(i)
            if shifted_elem < 0 and shape_a[i] != 1:
                broadcasted_dims_2.add(shifted_elem)

        for i in range(len(shape_b)):
            shifted_elem = i + shape_a_size_diff
            if shifted_elem >= 0 and shape_b[i] == 1 and shape_a[shifted_elem] != 1:
                broadcasted_dims_2.add(i)
            if shifted_elem < 0 and shape_b[i] != 1:
                broadcasted_dims_1.add(shifted_elem)

        return bool(broadcasted_dims_1) and bool(broadcasted_dims_2)


class GatherPruningOp(BasePruningOp):
    @classmethod
    def mask_propagation(
        cls, node: NNCFNode, graph: NNCFGraph, tensor_processor: Type[NNCFPruningBaseTensorProcessor]
    ) -> None:
        input_masks = get_input_masks(node, graph)
        assert len(input_masks) == 1
        input_mask = input_masks[0]
        node.attributes["output_mask"] = cls._get_output_mask(input_mask, node, graph)

    @classmethod
    def _get_output_mask(
        cls, input_mask: Optional[PropagationMask], node: NNCFNode, graph: NNCFGraph
    ) -> PropagationMask:
        if not input_mask:
            return None
        output_mask = PropagationMask()
        removed_axis = cls._is_dim_removed_by_splitting(graph, node)
        if removed_axis is not None:
            for dim, groups in input_mask.dim_groups_map.items():
                if dim != removed_axis:
                    shifted_dim = dim - 1
                    output_mask.dim_groups_map[shifted_dim] = groups
                    # other groups propagated further
                else:
                    for group in groups:
                        group.invalidate()
        else:
            input_mask.invalidate_groups()
        return output_mask

    @classmethod
    def _is_dim_removed_by_splitting(cls, graph: NNCFGraph, node: NNCFNode) -> Optional[int]:
        """
        Determines whether the operations going from parent's op is equivalent to split
        Currently, limited to the case of simple __getitem__ with axis=0 and integer key rather than more general
        `gather` operation or __getitem__ with slice or tuple.
        q, k, v = qkv[0], qkv[1], qkv[2] (like in official SwinTransformer implementation)

        - look at all consumer of parent node
        - all of them should
            - be getitem
            - slice along the same axis
            - combine the whole input (the keys should contain all dimension along axis. size of first dim=3 => [0,1,2])
        :param graph: NNCF graph to work with.
        :param node: A child node with __getitem__ operation.
        :return : axis that is removed by split, or None otherwise
        """
        split_axis = None
        if isinstance(node.layer_attributes, GetItemLayerAttributes):
            input_edge = graph.get_input_edges(node)[0]
            input_shape = input_edge.tensor_shape
            parent_node = input_edge.from_node
            child_nodes = graph.get_next_nodes(parent_node)
            child_attributes = [cnode.layer_attributes for cnode in child_nodes]
            all_getitem = all(isinstance(ca, GetItemLayerAttributes) for ca in child_attributes)
            assert all_getitem, "currently supported only case with all  __getitem__ on branches"
            all_int_keys = all(isinstance(ca.key, int) for ca in child_attributes)
            # currently supported only case __getitem__ with single int, no slices
            if not all_int_keys:
                return None
            all_keys = set(ca.key for ca in child_attributes)
            split_dim = input_shape[0]
            if all_keys == set(range(split_dim)):
                split_axis = 0
        return split_axis


class SplitPruningOp(BasePruningOp):
    @classmethod
    def mask_propagation(
        cls, node: NNCFNode, graph: NNCFGraph, tensor_processor: Type[NNCFPruningBaseTensorProcessor]
    ) -> None:
        input_masks = get_input_masks(node, graph)
        output_mask = None
        if input_masks:
            assert len(input_masks) == 1

            assert isinstance(node.layer_attributes, MultipleOutputLayerAttributes)
            chunk_axis = node.layer_attributes.axis
            chunks = node.layer_attributes.chunks

            input_edge = graph.get_input_edges(node)[0]
            output_edge = graph.get_output_edges(node)[0]
            input_shape = input_edge.tensor_shape
            output_shape = output_edge.tensor_shape
            is_chunk_axis_removed = (chunks == input_shape[chunk_axis]) and (len(input_shape) > len(output_shape))
            is_unit_chunk = input_shape[chunk_axis] // chunks == 1
            if is_chunk_axis_removed or is_unit_chunk:
                output_mask = PropagationMask()
                for dim, groups in input_masks[0].dim_groups_map.items():
                    if dim != chunk_axis:
                        # not affected by split groups are propagated further
                        output_mask.dim_groups_map[dim] = groups
                    else:
                        # invalidate groups, that assigned to the removed dimension or
                        # to the dimension that have 1 channel
                        for group in groups:
                            group.invalidate()
            else:
                nncf_logger.warning(
                    "symbolic mask propagation for split by prune dimension is not implemented, "
                    "just propagate further for now"
                )
        node.attributes["output_mask"] = output_mask


class ReshapeMode(Enum):
    """
    Defines the mode of reshape.
    Here's examples of reshaping for each mode:
        Extend: [N,C,H,W] ----(reshaped to)---> [N, C1, C2, H, W1, W2], when C=C1*C2 and W=W1*W2
        Shrink is opposite to Extend: [N, C1, C2, H, W1, W2] ----(reshaped to)---> [N,C,H,W]
        "Identity without ones" happens when removing ones from input and output shapes leads to identity mapping.
            For instance: [1, N] -> [N] is [N] or [N, 1, C] -> [1, N, 1, 1, C]
        Default - all other cases.
    """

    SHRINK = auto()
    EXTEND = auto()
    DEFAULT = auto()
    IDENTITY_WITHOUT_ONES = auto()


class ReshapePruningOp(BasePruningOp):
    DIMENSION_MAP = Dict[int, List[int]]

    @classmethod
    def mask_propagation(
        cls, node: NNCFNode, graph: NNCFGraph, tensor_processor: Type[NNCFPruningBaseTensorProcessor]
    ) -> None:
        input_masks = get_input_masks(node, graph)
        assert len(input_masks) == 1
        input_mask = input_masks[0]
        output_mask = PropagationMask()
        if input_mask and node.layer_attributes:
            in_map, _, mode = cls.parse_reshape(node.layer_attributes)
            input_shape = node.layer_attributes.input_shape
            output_shape = node.layer_attributes.output_shape

            if mode == ReshapeMode.DEFAULT:
                nncf_logger.warning(
                    f"Detected not supported reshape from {input_shape} to {output_shape}. "
                    f"The propagation mask is invalidated for node={node.node_name}."
                )
                input_mask.invalidate_groups()
            elif mode == ReshapeMode.IDENTITY_WITHOUT_ONES:
                # pruning dimension is not shrunk and extended, just assign a new location in the output
                for dim, groups in input_mask.dim_groups_map.items():
                    assert len(in_map[dim]) == 1, "assume a mapping to single int by definition of identity"
                    shifted_dim = in_map[dim][0]
                    output_mask.dim_groups_map[shifted_dim] = groups
            elif mode == ReshapeMode.EXTEND:
                for dim, groups in input_mask.dim_groups_map.items():
                    if len(groups) > 1:
                        raise NotImplementedError("Extend reshape for several groups is not supported yet")
                    if len(in_map[dim]) == 1:
                        # pruning dimension is not extended, just assign a new location in the output
                        shifted_dim = in_map[dim][0]
                        output_mask.dim_groups_map[shifted_dim] = groups
                    else:
                        # extension like this: S -> [A,B,C,D]. Need to split groups.
                        input_channels = input_shape[dim]
                        list_output_channels = [output_shape[x] for x in in_map[dim]]
                        group = groups[0]
                        if group.has_children():
                            raise NotImplementedError("Splitting BlockGroup with children is not implemented yet")
                        child_groups = cls._split_group(group, input_channels, list_output_channels)
                        for child_group, in_dim in zip(child_groups, in_map[dim]):
                            output_mask.dim_groups_map[in_dim] = [child_group]
            elif mode == ReshapeMode.SHRINK:
                # shrinking like: [A,B,C,D] -> S, just combine all groups under the same dimension
                grouping = defaultdict(list)
                for in_idx, groups in input_mask.dim_groups_map.items():
                    assert len(in_map[in_idx]) == 1, "assume a mapping to single int by definition of shrink"
                    grouping[in_map[in_idx][0]].extend(groups)
                output_mask.dim_groups_map = dict(grouping)

        node.attributes["output_mask"] = output_mask

    @staticmethod
    def _can_reach_number_by_multiply(number_to_reach: int, array: List[int], start_idx: int) -> Tuple[bool, int]:
        """
        Sequentially multiplies elements in the given array starting from the given start index until reaching or
        exceeding the given number.

        :param number_to_reach: number to reach while multiplying elements.
        :param array: array of numbers.
        :param start_idx: the index, starting from which the numbers in the given array will be sequentially multiplied.
        :return: a pair, when first element is True when the given number is reached, False - otherwise.
            The second element is the index of last element in the array when the product of numbers reached or
            exceeded the given number.
        """
        idx = start_idx
        accum = array[idx]
        while accum < number_to_reach:
            idx += 1
            accum *= array[idx]
        if accum > number_to_reach:
            return (False, idx)
        return (True, idx)

    @classmethod
    def _map_dims_(
        cls,
        source_array: List[int],
        target_array: List[int],
        source_idx: int,
        start_target_idx: int,
        source_to_target_map: DIMENSION_MAP,
        target_to_source_map: DIMENSION_MAP,
    ) -> Tuple[bool, int]:
        """
        The function tries to decompose one element in the source array by the product of sequential elements in the
        target array. If it can be decomposed it's considered "mapped". i.e. the source element corresponds to the list
        of elements in the target array, end vice-versa, elements in the target array corresponds to the element in the
        source array. By this correspondence, mappings of element's indexes from source to target and from target to
        source are built.

        :param source_array: array to get a single element for mapping to elements in the target array.
        :param target_array: array to find element(s) that map(s) to a single element in the source array.
        :param source_idx: index of the single element in the source array.
        :param target_start_idx: index in the target array to start sequentially multiplying elements until reaching
            the element from source array.
        :param source_to_target_map: mapping of indexes of source elements to indexes of target elements.
        :param target_to_source_map: mapping of indexes of target elements to indexes of source elements.
        :return: a pair, when first element is True when the element from source array can be mapped to sequential
            element(s) of target array, False - otherwise.
            The second element of pair is the index of last mapped element in the target array.
        """
        res, last_target_index = cls._can_reach_number_by_multiply(
            number_to_reach=source_array[source_idx], array=target_array, start_idx=start_target_idx
        )
        if not res:
            return (res, last_target_index)
        source_to_target_map[source_idx] = list(range(start_target_idx, last_target_index + 1))
        for idx in range(start_target_idx, last_target_index + 1):
            target_to_source_map[idx] = [source_idx]
        return (res, last_target_index)

    @classmethod
    def parse_reshape(
        cls, reshape_attributes: ReshapeLayerAttributes
    ) -> Tuple[DIMENSION_MAP, DIMENSION_MAP, ReshapeMode]:
        """
        Parse reshape semantic by mapping elements in the input shape to elements of output shape.

        :param reshape_attributes: attributes that contains input and output shape.
        :return: return two mappings (from input to output, and vice versa) and reshape mode.
        """
        input_shape_not_cut = reshape_attributes.input_shape
        output_shape_not_cut = reshape_attributes.output_shape

        in_indexes_not_cut_map = [i for i, dim in enumerate(input_shape_not_cut) if dim != 1]
        out_indexes_not_cut_map = [i for i, dim in enumerate(output_shape_not_cut) if dim != 1]

        input_shape = list(filter(lambda x: x != 1, input_shape_not_cut))
        output_shape = list(filter(lambda x: x != 1, output_shape_not_cut))

        in_idx = 0
        out_idx = 0
        in_map = {}
        out_map = {}

        mode = ReshapeMode.DEFAULT

        while in_idx < len(input_shape) and out_idx < len(output_shape):
            if input_shape[in_idx] == output_shape[out_idx]:
                in_map[in_idx] = [out_idx]
                out_map[out_idx] = [in_idx]
            elif input_shape[in_idx] > output_shape[out_idx]:
                res, out_idx = cls._map_dims_(
                    source_array=input_shape,
                    target_array=output_shape,
                    source_idx=in_idx,
                    start_target_idx=out_idx,
                    source_to_target_map=in_map,
                    target_to_source_map=out_map,
                )
                if not res or mode == ReshapeMode.SHRINK:
                    return None, None, ReshapeMode.DEFAULT
                mode = ReshapeMode.EXTEND
            else:
                res, in_idx = cls._map_dims_(
                    source_array=output_shape,
                    target_array=input_shape,
                    source_idx=out_idx,
                    start_target_idx=in_idx,
                    source_to_target_map=out_map,
                    target_to_source_map=in_map,
                )
                if not res or mode == ReshapeMode.EXTEND:
                    return None, None, ReshapeMode.DEFAULT
                mode = ReshapeMode.SHRINK
            in_idx += 1
            out_idx += 1

        if mode == ReshapeMode.DEFAULT:
            mode = ReshapeMode.IDENTITY_WITHOUT_ONES

        in_map_not_cut = cls._convert_to_not_cut(in_indexes_not_cut_map, out_indexes_not_cut_map, in_map)
        out_map_not_cut = cls._convert_to_not_cut(out_indexes_not_cut_map, in_indexes_not_cut_map, out_map)
        return in_map_not_cut, out_map_not_cut, mode

    @classmethod
    def _convert_to_not_cut(
        cls,
        source_indexes_not_cut_map: List[int],
        target_indexes_not_cut_map: List[int],
        source_to_targets_map: DIMENSION_MAP,
    ) -> DIMENSION_MAP:
        """
        Perform re-mapping indexes for the cut arrays to non-cut ones.

        :param source_indexes_not_cut_map: indexes of not cut elements in the uncut source array.
            The indexes are in the order of the elements in the array, so it can be used to map indexes from a cut array
            to uncut one.
        :param target_indexes_not_cut_map: indexes of not cut elements in the uncut target array. The indexes are in the
            order of the elements in the array, so it can be used to map indexes from a cut array to uncut one.
        :param source_to_targets_map: mapping from source elements to target ones in the cut arrays.
        :return: mapping from source elements to target ones in the not cut arrays.
        """
        source_to_targets_map_not_cut = {}
        for source_index, target_indexes in source_to_targets_map.items():
            source_index_not_cut = source_indexes_not_cut_map[source_index]
            target_indexes_not_cut = list(map(lambda x: target_indexes_not_cut_map[x], target_indexes))
            source_to_targets_map_not_cut[source_index_not_cut] = target_indexes_not_cut
        return source_to_targets_map_not_cut

    @staticmethod
    def _split_block(block: PruningBlock, list_output_channels: List[int]) -> List[PruningBlock]:
        """
        Splits a pruning block into multiple blocks.
        It's applied when some number of channels S is reshaped to N channels, e.g. S -> [A,B,C,D] and S=A*B*C*D.
        Now we assume that pruning is possible along each new dimension instead of S one.
        The pruning block encoding pruning of a single channel from S is no longer valid.
        This function creates new blocks that encodes how many channels from S is pruned if we prune along a new
        dimension from [A,B,C,D].
        It forms new constraints by the following rules:
            | size  | offset          |
            |-------|-----------------|
            | 1     | D % S           |
            | D     | C*D % S         |
            | C*D   | B*C*D % S       |
            | B*C*D | A*B*C*D % S = 0 |

        :param block: pruning block to split
        :param list_output_channels: list of output channels (A,B,C,D)
        :return:
        """
        if len(list_output_channels) == 1:
            raise RuntimeError

        dot_product = reduce((lambda x, y: x * y), list_output_channels)

        current_size = dot_product
        new_blocks = []
        divided_shapes = filter(lambda x: x != 1, list_output_channels)
        for divided_shape in divided_shapes:
            offset = int(current_size % dot_product)
            current_size /= divided_shape
            new_block = copy.copy(block)
            new_block.size = int(current_size)
            new_block.offset = offset
            new_blocks.append(new_block)
        return new_blocks

    @classmethod
    def _split_group(
        cls, group: PropagationGroup, input_channels: int, list_output_channels: List[int]
    ) -> List[PropagationGroup]:
        """
        Splits a pruning block within the given group into multiple blocks - each within new groups.
        It's applied when some number of input channels S is reshaped to N output channels,
        e.g. S -> [A,B,C,D] and S=A*B*C*D.

        :param input_channels: splitted input channels (S).
        :param list_output_channels: list of output channels (A,B,C,D).
        :param group: a group for splitting.
        :return: list of new groups which results from the split.
        """
        dot_product = reduce((lambda x, y: x * y), list_output_channels)
        assert dot_product == input_channels

        new_groups: List[PropagationGroup] = []
        new_blocks = cls._split_block(group.block, list_output_channels)
        for block in new_blocks:
            new_group = PropagationGroup(block=block, producers=group.get_producers(), consumers=group.get_consumers())
            group.add_child(new_group)
            new_groups.append(new_group)
        return new_groups


class TransposePruningOp(BasePruningOp):
    @classmethod
    def mask_propagation(
        cls, node: NNCFNode, graph: NNCFGraph, tensor_processor: Type[NNCFPruningBaseTensorProcessor]
    ) -> None:
        input_masks = get_input_masks(node, graph)
        assert len(input_masks) == 1
        input_mask = input_masks[0]
        if not input_mask:
            node.attributes["output_mask"] = None
            return

        if isinstance(node.layer_attributes, TransposeLayerAttributes):
            input_tensors_shapes = [x.tensor_shape for x in graph.get_input_edges(node)]
            assert len(input_tensors_shapes) == 1
            new_order = list(range(len(input_tensors_shapes[0])))
            dim0 = node.layer_attributes.dim0
            dim1 = node.layer_attributes.dim1
            new_order[dim1], new_order[dim0] = new_order[dim0], new_order[dim1]
        elif isinstance(node.layer_attributes, PermuteLayerAttributes):
            new_order = node.layer_attributes.permutation

        idx_map = [
            (old_idx, new_idx) for new_idx, old_idx in enumerate(new_order) if old_idx in input_mask.dim_groups_map
        ]
        output_mask = PropagationMask(
            dim_groups_map={new_idx: input_mask.dim_groups_map[old_idx] for old_idx, new_idx in idx_map}
        )

        node.attributes["output_mask"] = output_mask


class StopMaskForwardPruningOp(BasePruningOp):
    @classmethod
    def mask_propagation(
        cls, node: NNCFNode, graph: NNCFGraph, tensor_processor: Type[NNCFPruningBaseTensorProcessor]
    ) -> None:
        input_masks = get_input_masks(node, graph)
        cls.invalidate_masks(input_masks)
        node.attributes["output_mask"] = None


class ExpandAsPruningOp(BasePruningOp):
    @classmethod
    def mask_propagation(
        cls, node: NNCFNode, graph: NNCFGraph, tensor_processor: Type[NNCFPruningBaseTensorProcessor]
    ) -> None:
        input_edges = graph.get_input_edges(node)
        assert len(input_edges) == 2, "expand should always have 2 inputs"
        input_to_expand = input_edges[0].from_node
        mask = input_to_expand.attributes.get("output_mask")
        propagated_mask = None
        if mask:
            nncf_logger.warning(
                "expand_as is applied to the node with propagation mask. Currently, it's not supported "
                f"and mask is invalidated. node_name={node.node_name}"
            )
            mask.invalidate_groups()
        input_to_get_shape = input_edges[1].from_node
        mask: PropagationMask = input_to_get_shape.attributes.get("output_mask")
        if mask:
            target_shape = input_edges[1].tensor_shape
            source_shape = input_edges[0].tensor_shape
            for dim, groups in mask.dim_groups_map.items():
                if target_shape[dim] == source_shape[dim]:
                    nncf_logger.warning(
                        "expand_as takes the shape from the node with propagation mask and pruning "
                        "dimension in the mask matches the dimension in the expanded input. Currently, "
                        f"it's not supported and mask is invalidated. node_name={node.node_name}"
                    )
                    for group in groups:
                        group.invalidate()
            # TODO: (nlyalyus) assume that expand_as is on constant path that does not affect pruning, otherwise pruning
            # of self attention block would be not possible in the general case.
            propagated_mask = mask
        node.attributes["output_mask"] = propagated_mask


class ScatterPruningOp(BasePruningOp):
    @classmethod
    def mask_propagation(
        cls, node: NNCFNode, graph: NNCFGraph, tensor_processor: Type[NNCFPruningBaseTensorProcessor]
    ) -> None:
        input_masks = [input_edge.from_node.attributes.get("output_mask") for input_edge in graph.get_input_edges(node)]
        assert len(input_masks) == 2, "expect that masked_fill should always have 2 inputs"
        i1, i2 = input_masks
        propagated_mask = None
        if i1 != i2:
            nncf_logger.warning(
                "expand_as takes the shape from the node with propagation mask and pruning "
                "dimension in the mask matches the dimension in the expanded input. Currently, "
                f"it's not supported and mask is invalidated. node_name={node.node_name}"
            )
            if i1:
                i1.invalidate_groups()
        else:
            propagated_mask = i1
        node.attributes["output_mask"] = propagated_mask



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from nncf.common.graph.operator_metatypes import UnknownMetatype
from nncf.common.pruning.utils import PruningOperationsMetatypeRegistry
from nncf.experimental.common.pruning.operations import BatchNormPruningOp
from nncf.experimental.common.pruning.operations import ElementwisePruningOp
from nncf.experimental.common.pruning.operations import ExpandAsPruningOp
from nncf.experimental.common.pruning.operations import GatherPruningOp
from nncf.experimental.common.pruning.operations import GroupNormPruningOp
from nncf.experimental.common.pruning.operations import IdentityMaskForwardPruningOp
from nncf.experimental.common.pruning.operations import InputPruningOp
from nncf.experimental.common.pruning.operations import LinearPruningOp
from nncf.experimental.common.pruning.operations import OutputPruningOp
from nncf.experimental.common.pruning.operations import ReshapePruningOp
from nncf.experimental.common.pruning.operations import ScatterPruningOp
from nncf.experimental.common.pruning.operations import SplitPruningOp
from nncf.experimental.common.pruning.operations import StopMaskForwardPruningOp
from nncf.experimental.common.pruning.operations import TransposePruningOp
from nncf.torch.graph.operator_metatypes import PTAdaptiveMaxPool2dMetatype
from nncf.torch.graph.operator_metatypes import PTAddMetatype
from nncf.torch.graph.operator_metatypes import PTAvgPool2dMetatype
from nncf.torch.graph.operator_metatypes import PTBatchNormMetatype
from nncf.torch.graph.operator_metatypes import PTDivMetatype
from nncf.torch.graph.operator_metatypes import PTDropoutMetatype
from nncf.torch.graph.operator_metatypes import PTELUMetatype
from nncf.torch.graph.operator_metatypes import PTExpandAsMetatype
from nncf.torch.graph.operator_metatypes import PTGatherMetatype
from nncf.torch.graph.operator_metatypes import PTGELUMetatype
from nncf.torch.graph.operator_metatypes import PTGroupNormMetatype
from nncf.torch.graph.operator_metatypes import PTHardSigmoidMetatype
from nncf.torch.graph.operator_metatypes import PTHardSwishMetatype
from nncf.torch.graph.operator_metatypes import PTHardTanhMetatype
from nncf.torch.graph.operator_metatypes import PTInputNoopMetatype
from nncf.torch.graph.operator_metatypes import PTLeakyRELUMetatype
from nncf.torch.graph.operator_metatypes import PTLinearMetatype
from nncf.torch.graph.operator_metatypes import PTMatMulMetatype
from nncf.torch.graph.operator_metatypes import PTMaxMetatype
from nncf.torch.graph.operator_metatypes import PTMaxPool2dMetatype
from nncf.torch.graph.operator_metatypes import PTMeanMetatype
from nncf.torch.graph.operator_metatypes import PTMinMetatype
from nncf.torch.graph.operator_metatypes import PTMulMetatype
from nncf.torch.graph.operator_metatypes import PTNoopMetatype
from nncf.torch.graph.operator_metatypes import PTOutputNoopMetatype
from nncf.torch.graph.operator_metatypes import PTPowerMetatype
from nncf.torch.graph.operator_metatypes import PTPRELUMetatype
from nncf.torch.graph.operator_metatypes import PTRELU6Metatype
from nncf.torch.graph.operator_metatypes import PTRELUMetatype
from nncf.torch.graph.operator_metatypes import PTReshapeMetatype
from nncf.torch.graph.operator_metatypes import PTScatterMetatype
from nncf.torch.graph.operator_metatypes import PTSigmoidMetatype
from nncf.torch.graph.operator_metatypes import PTSILUMetatype
from nncf.torch.graph.operator_metatypes import PTSoftmaxMetatype
from nncf.torch.graph.operator_metatypes import PTSplitMetatype
from nncf.torch.graph.operator_metatypes import PTSqueezeMetatype
from nncf.torch.graph.operator_metatypes import PTSubMetatype
from nncf.torch.graph.operator_metatypes import PTSumMetatype
from nncf.torch.graph.operator_metatypes import PTTanhMetatype
from nncf.torch.graph.operator_metatypes import PTTransposeMetatype

PT_EXPERIMENTAL_PRUNING_OPERATOR_METATYPES = PruningOperationsMetatypeRegistry("operator_metatypes")


@PT_EXPERIMENTAL_PRUNING_OPERATOR_METATYPES.register("model_input")
class PTInputPruningOp(InputPruningOp):
    subtypes = [PTInputNoopMetatype]


@PT_EXPERIMENTAL_PRUNING_OPERATOR_METATYPES.register("model_output")
class PTOutputPruningOp(OutputPruningOp):
    subtypes = [PTOutputNoopMetatype]


@PT_EXPERIMENTAL_PRUNING_OPERATOR_METATYPES.register("identity_mask_propagation")
class PTIdentityMaskForwardPruningOp(IdentityMaskForwardPruningOp):
    subtypes = [
        PTHardTanhMetatype,
        PTTanhMetatype,
        PTRELUMetatype,
        PTRELU6Metatype,
        PTLeakyRELUMetatype,
        PTPRELUMetatype,
        PTELUMetatype,
        PTGELUMetatype,
        PTSigmoidMetatype,
        PTSoftmaxMetatype,
        PTAvgPool2dMetatype,
        PTAdaptiveMaxPool2dMetatype,
        PTMaxPool2dMetatype,
        PTMeanMetatype,
        PTDropoutMetatype,
        PTSILUMetatype,
        PTPowerMetatype,
        PTHardSwishMetatype,
        PTHardSigmoidMetatype,
        PTNoopMetatype,
    ]
    additional_types = ["h_sigmoid", "h_swish", "RELU"]


@PT_EXPERIMENTAL_PRUNING_OPERATOR_METATYPES.register("linear")
class PTLinearPruningOp(LinearPruningOp):
    subtypes = [PTLinearMetatype, PTMatMulMetatype]


@PT_EXPERIMENTAL_PRUNING_OPERATOR_METATYPES.register("batch_norm")
class PTBatchNormPruningOp(BatchNormPruningOp):
    subtypes = [PTBatchNormMetatype]


@PT_EXPERIMENTAL_PRUNING_OPERATOR_METATYPES.register("group_norm")
class PTGroupNormPruningOp(GroupNormPruningOp):
    subtypes = [PTGroupNormMetatype]


@PT_EXPERIMENTAL_PRUNING_OPERATOR_METATYPES.register("elementwise")
class PTElementwisePruningOp(ElementwisePruningOp):
    subtypes = [PTAddMetatype, PTSubMetatype, PTDivMetatype, PTMulMetatype]


@PT_EXPERIMENTAL_PRUNING_OPERATOR_METATYPES.register("stop_propagation_ops")
class PTStopMaskForwardPruningOp(StopMaskForwardPruningOp):
    subtypes = [PTMaxMetatype, PTMinMetatype, PTSumMetatype, UnknownMetatype]


@PT_EXPERIMENTAL_PRUNING_OPERATOR_METATYPES.register("transpose")
class PTTransposePruningOp(TransposePruningOp):
    subtypes = [PTTransposeMetatype]


@PT_EXPERIMENTAL_PRUNING_OPERATOR_METATYPES.register("reshape")
class PTReshape(ReshapePruningOp):
    subtypes = [PTReshapeMetatype, PTSqueezeMetatype]


@PT_EXPERIMENTAL_PRUNING_OPERATOR_METATYPES.register("split")
class PTSplitPruningOp(SplitPruningOp):
    subtypes = [PTSplitMetatype]


@PT_EXPERIMENTAL_PRUNING_OPERATOR_METATYPES.register("gather")
class PTGatherPruningOp(GatherPruningOp):
    subtypes = [PTGatherMetatype]


@PT_EXPERIMENTAL_PRUNING_OPERATOR_METATYPES.register("expend_as")
class PTExpandAsPruningOp(ExpandAsPruningOp):
    subtypes = [PTExpandAsMetatype]


@PT_EXPERIMENTAL_PRUNING_OPERATOR_METATYPES.register("masked_fill")
class PTScatterPruningOp(ScatterPruningOp):
    subtypes = [PTScatterMetatype]



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Deque, List, Optional, Tuple, Union

import numpy as np

from nncf.common.tensor import TensorElementsType
from nncf.common.tensor_statistics.collectors import NNCFCollectorTensorProcessor
from nncf.experimental.common.tensor_statistics.collectors import AbsMaxReducer
from nncf.experimental.common.tensor_statistics.collectors import AbsQuantileReducer
from nncf.experimental.common.tensor_statistics.collectors import BatchMeanReducer
from nncf.experimental.common.tensor_statistics.collectors import MaxReducer
from nncf.experimental.common.tensor_statistics.collectors import MeanAggregator
from nncf.experimental.common.tensor_statistics.collectors import MeanPerChReducer
from nncf.experimental.common.tensor_statistics.collectors import MeanReducer
from nncf.experimental.common.tensor_statistics.collectors import MinReducer
from nncf.experimental.common.tensor_statistics.collectors import NoopAggregator
from nncf.experimental.common.tensor_statistics.collectors import NoopReducer
from nncf.experimental.common.tensor_statistics.collectors import QuantileReducer
from nncf.experimental.common.tensor_statistics.collectors import RawReducer
from nncf.experimental.common.tensor_statistics.collectors import ShapeAggregator
from nncf.experimental.common.tensor_statistics.collectors import TensorCollector
from nncf.experimental.common.tensor_statistics.collectors import TensorReducerBase
from nncf.onnx.statistics.statistics import ONNXMeanTensorStatistic
from nncf.onnx.statistics.statistics import ONNXRawTensorStatistic
from nncf.onnx.tensor import ONNXNNCFTensor
from nncf.quantization.advanced_parameters import StatisticsType


class ONNXNNCFCollectorTensorProcessor(NNCFCollectorTensorProcessor):
    """
    A realization of the processing methods for ONNXNNCFTensors.
    """

    @staticmethod
    def reduce_min(
        x: ONNXNNCFTensor, axis: Union[int, Tuple[int, ...], List[int]], keepdims: bool = False
    ) -> ONNXNNCFTensor:
        return ONNXNNCFTensor(np.amin(x.tensor, axis=axis, keepdims=keepdims))

    @staticmethod
    def reduce_max(
        x: ONNXNNCFTensor, axis: Union[int, Tuple[int, ...], List[int]], keepdims: bool = False
    ) -> ONNXNNCFTensor:
        return ONNXNNCFTensor(np.amax(x.tensor, axis=axis, keepdims=keepdims))

    @staticmethod
    def abs(x: ONNXNNCFTensor) -> ONNXNNCFTensor:
        return ONNXNNCFTensor(np.abs(x.tensor))

    @staticmethod
    def min(x1: ONNXNNCFTensor, x2: ONNXNNCFTensor) -> ONNXNNCFTensor:
        return ONNXNNCFTensor(np.minimum(x1.tensor, x2.tensor))

    @staticmethod
    def max(x1: ONNXNNCFTensor, x2: ONNXNNCFTensor) -> ONNXNNCFTensor:
        return ONNXNNCFTensor(np.maximum(x1.tensor, x2.tensor))

    @staticmethod
    def mean(x: ONNXNNCFTensor, axis: Union[int, Tuple[int, ...], List[int]], keepdims: bool = False) -> ONNXNNCFTensor:
        return ONNXNNCFTensor(np.mean(x.tensor, axis=axis, keepdims=keepdims))

    @staticmethod
    def median(
        x: ONNXNNCFTensor, axis: Union[int, Tuple[int, ...], List[int]], keepdims: bool = False
    ) -> ONNXNNCFTensor:
        return ONNXNNCFTensor(np.median(x.tensor, axis=axis, keepdims=keepdims))

    @classmethod
    def masked_mean(
        cls,
        x: ONNXNNCFTensor,
        axis: Optional[Union[int, Tuple[int, ...], List[int]]],
        mask: Optional[ONNXNNCFTensor],
        keepdims: bool = False,
    ) -> ONNXNNCFTensor:
        if mask is None:
            return cls.mean(x, axis=axis, keepdims=keepdims)
        masked_x = np.ma.array(x.tensor, mask=mask.tensor)
        result = np.ma.mean(masked_x, axis=axis, keepdims=keepdims)
        if isinstance(result, np.ma.MaskedArray):
            return ONNXNNCFTensor(result.data)
        return ONNXNNCFTensor(result)

    @classmethod
    def masked_median(
        cls,
        x: ONNXNNCFTensor,
        axis: Optional[Union[int, Tuple[int, ...], List[int]]],
        mask: Optional[ONNXNNCFTensor],
        keepdims: bool = False,
    ) -> ONNXNNCFTensor:
        if mask is None:
            return cls.median(x, axis=axis, keepdims=keepdims)
        masked_x = np.ma.array(x.tensor, mask=mask.tensor)
        result = np.ma.median(masked_x, axis=axis, keepdims=keepdims)
        if isinstance(result, np.ma.MaskedArray):
            return ONNXNNCFTensor(result.data)
        return ONNXNNCFTensor(result)

    @staticmethod
    def logical_or(input_: ONNXNNCFTensor, other: ONNXNNCFTensor) -> ONNXNNCFTensor:
        return ONNXNNCFTensor(np.logical_or(input_.tensor, other.tensor))

    @staticmethod
    def less(input_: ONNXNNCFTensor, other: ONNXNNCFTensor) -> ONNXNNCFTensor:
        return ONNXNNCFTensor(input_.tensor < other.tensor)

    @staticmethod
    def stack(x: Union[List[ONNXNNCFTensor], Deque[ONNXNNCFTensor]], axis: int = 0) -> ONNXNNCFTensor:
        x = [t.tensor for t in x]
        return ONNXNNCFTensor(np.stack(x, axis=axis))

    @staticmethod
    def unstack(x: ONNXNNCFTensor, axis: int = 0) -> List[ONNXNNCFTensor]:
        return [ONNXNNCFTensor(np.squeeze(e, axis)) for e in np.split(x.tensor, x.tensor.shape[axis], axis=axis)]

    @staticmethod
    def squeeze(x: ONNXNNCFTensor, dim: Optional[Union[int, Tuple[int, ...]]] = None) -> ONNXNNCFTensor:
        return ONNXNNCFTensor(np.squeeze(x.tensor, axis=dim))

    @staticmethod
    def sum(tensor: ONNXNNCFTensor) -> ONNXNNCFTensor:
        return ONNXNNCFTensor(np.sum(tensor.tensor))

    @staticmethod
    def quantile(
        tensor: ONNXNNCFTensor,
        quantile: Union[float, List[float]],
        axis: Union[int, Tuple[int, ...], List[int]],
        keepdims: bool = False,
    ) -> List[ONNXNNCFTensor]:
        result = np.quantile(tensor.tensor, quantile, axis, keepdims=keepdims)
        return [ONNXNNCFTensor(x) for x in result]

    @classmethod
    def percentile(
        cls,
        tensor: ONNXNNCFTensor,
        percentile: Union[float, List[float]],
        axis: Union[int, Tuple[int, ...], List[int]],
        keepdims: bool = False,
    ) -> List[TensorElementsType]:
        quantile = np.true_divide(percentile, 100)
        return cls.quantile(tensor, quantile=quantile, axis=axis, keepdims=keepdims)

    @staticmethod
    def mean_per_channel(x: ONNXNNCFTensor, axis: int) -> ONNXNNCFTensor:
        if len(x.shape) < 3:
            return ONNXNNCFTensor(np.mean(x.tensor, axis=0))
        x = np.moveaxis(x.tensor, axis, 1)
        t = x.reshape(x.shape[0], x.shape[1], -1)
        return ONNXNNCFTensor(np.mean(t, axis=(0, 2)))

    @staticmethod
    def transpose(x: ONNXNNCFTensor, axes: Tuple[int, ...]) -> ONNXNNCFTensor:
        return ONNXNNCFTensor(np.transpose(x.tensor, axes))

    @staticmethod
    def reshape(x: ONNXNNCFTensor, shape: Tuple[int, ...]) -> ONNXNNCFTensor:
        return ONNXNNCFTensor(np.reshape(x.tensor, shape))

    @staticmethod
    def cat(x: List[ONNXNNCFTensor], axis: int) -> ONNXNNCFTensor:
        x = [t.tensor for t in x]
        return ONNXNNCFTensor(np.concatenate(x, axis))

    @staticmethod
    def batch_mean(x: ONNXNNCFTensor) -> ONNXNNCFTensor:
        return ONNXNNCFTensor(np.mean(x.tensor, axis=0, keepdims=True))

    @staticmethod
    def sub(a: ONNXNNCFTensor, b: ONNXNNCFTensor) -> ONNXNNCFTensor:
        return ONNXNNCFTensor(a.tensor - b.tensor)

    @staticmethod
    def zero_elements(x: ONNXNNCFTensor) -> ONNXNNCFTensor:
        np_tensor = x.tensor
        eps = np.finfo(np_tensor.dtype).eps
        return ONNXNNCFTensor(np.abs(np_tensor) < eps)


class ONNXBasicReducer(TensorReducerBase):
    def _get_processor(self):
        return ONNXNNCFCollectorTensorProcessor

    def get_inplace_fn(self):
        raise NotImplementedError("ONNX backend has no support of inplace statistics yet.")


class ONNXMinReducer(ONNXBasicReducer, MinReducer):
    pass


class ONNXMaxReducer(ONNXBasicReducer, MaxReducer):
    pass


class ONNXAbsMaxReducer(ONNXBasicReducer, AbsMaxReducer):
    pass


class ONNXMeanReducer(ONNXBasicReducer, MeanReducer):
    pass


class ONNXQuantileReducer(ONNXBasicReducer, QuantileReducer):
    pass


class ONNXAbsQuantileReducer(ONNXBasicReducer, AbsQuantileReducer):
    pass


class ONNXBatchMeanReducer(ONNXBasicReducer, BatchMeanReducer):
    pass


class ONNXMeanPerChanelReducer(ONNXBasicReducer, MeanPerChReducer):
    pass


def get_mean_statistic_collector(
    num_samples: int, channel_axis: int, window_size: Optional[int] = None, inplace: bool = True
) -> TensorCollector:
    """
    Mean statistic collector builder.

    :param num_samples: Maximum number of samples to collect.
    :param channel_axis: Channel axis to use during reduction phase.
    :param window_size: Number of samples from the end of the list of collected samples to aggregate.
        Aggregates all available collected statistics in case parameter is None.
    :param inplace: Whether the mean reducer should be calculated inplace or out of place.
    :return: Mean statistic collector.
    """
    inplace = False
    if channel_axis == 0:
        reducer = ONNXBatchMeanReducer(inplace)
    else:
        reducer = ONNXMeanPerChanelReducer(channel_axis=channel_axis, inplace=inplace)
    noop_reducer = NoopReducer()

    kwargs = {
        "tensor_processor": ONNXNNCFCollectorTensorProcessor,
        "num_samples": num_samples,
        "window_size": window_size,
    }

    aggregate_mean = MeanAggregator(**kwargs)
    aggregate_shape = ShapeAggregator()

    collector = TensorCollector(ONNXMeanTensorStatistic)
    collector.register_statistic_branch(ONNXMeanTensorStatistic.MEAN_STAT, reducer, aggregate_mean)
    collector.register_statistic_branch(ONNXMeanTensorStatistic.SHAPE_STAT, noop_reducer, aggregate_shape)
    return collector


def get_raw_stat_collector(num_samples: int) -> TensorCollector:
    """
    Raw statistic collector builder.

    :param num_samples: Maximum number of samples to collect.
    :return: Raw statistic collector.
    """
    reducer = RawReducer()
    aggregator = NoopAggregator(num_samples)

    collector = TensorCollector(ONNXRawTensorStatistic)
    collector.register_statistic_branch(ONNXRawTensorStatistic.VALUES_STATS, reducer, aggregator)
    return collector


ONNX_REDUCERS_MAP = {
    StatisticsType.MIN: ONNXMinReducer,
    StatisticsType.MAX: ONNXMaxReducer,
    StatisticsType.ABS_MAX: ONNXAbsMaxReducer,
    StatisticsType.MEAN: ONNXMeanReducer,
    StatisticsType.QUANTILE: ONNXQuantileReducer,
    StatisticsType.ABS_QUANTILE: ONNXAbsQuantileReducer,
}



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict

import numpy as np
import onnxruntime as rt

from nncf.common.engine import Engine


class ONNXEngine(Engine):
    """
    Engine for ONNX backend using ONNXRuntime to infer the model.
    """

    def __init__(self, model, **rt_session_options):
        self.input_names = set()
        rt_session_options["providers"] = ["CPUExecutionProvider"]
        serialized_model = model.SerializeToString()
        self.sess = rt.InferenceSession(serialized_model, **rt_session_options)

        for inp in self.sess.get_inputs():
            self.input_names.add(inp.name)

    def infer(self, input_data: Dict[str, np.ndarray]) -> Dict[str, np.ndarray]:
        """
        Runs model on the provided input via ONNXRuntime InferenceSession.
        Returns the dictionary of model outputs by node names.
        :param input_data: inputs for the model
        :return output_data: models outputs
        """
        output_tensors = self.sess.run([], {k: v for k, v in input_data.items() if k in self.input_names})
        model_outputs = self.sess.get_outputs()

        return {output.name: tensor for tensor, output in zip(output_tensors, model_outputs)}
